from copy import copy
from enum import Enum
from typing import List, Dict, Optional, cast, TypeVar

from sqlalchemy.engine import Dialect

from bach import SortColumn
from bach.series import Series
from bach.expression import Expression, WindowFunctionExpression, join_expressions, ConstValueExpression
from bach.sql_model import BachSqlModel
from sql_models.util import is_postgres, is_bigquery, DatabaseNotSupportedException, is_athena

G = TypeVar('G', bound='GroupBy')


class WindowFunction(Enum):
    FIRST_VALUE = 'first_value'
    LAST_VALUE = 'last_value'
    NTH_VALUE = 'nth_value'
    LEAD = 'lead'
    LAG = 'lag'

    RANK = 'rank'
    DENSE_RANK = 'dense_rank'
    PERCENT_RANK = 'percent_rank'
    CUME_DIST = 'cume_dist'
    NTILE = 'ntile'
    ROW_NUMBER = 'row_number'

    _BQ_FUNCTIONS_THAT_DO_NOT_SUPPORT_WINDOW_FRAME_CLAUSE = (
        RANK,
        DENSE_RANK,
        PERCENT_RANK,
        CUME_DIST,
        NTILE,
        ROW_NUMBER,
        LAG,
        LEAD,
    )

    def supports_window_frame_clause(self, dialect: Dialect) -> bool:
        if is_bigquery(dialect):
            return (
                self.value not in WindowFunction._BQ_FUNCTIONS_THAT_DO_NOT_SUPPORT_WINDOW_FRAME_CLAUSE.value
            )

        if is_athena(dialect) or is_postgres(dialect):
            return True
        raise DatabaseNotSupportedException(dialect)


class WindowFrameMode(Enum):
    """
    Class representing the frame mode in a Window
    """
    ROWS = 0
    RANGE = 1


class WindowFrameBoundary(Enum):
    """
    Class representing the frame boundaries in a Window
        """

    # Order is important here (see third restriction above)
    PRECEDING = 0
    CURRENT_ROW = 1
    FOLLOWING = 2

    def frame_clause(self, value: int = None) -> str:
        """
        Generate the frame boundary sub-string
        """
        if self == self.CURRENT_ROW:
            if value is not None:
                raise ValueError('Value not supported with CURRENT ROW')
            return 'CURRENT ROW'
        else:
            if value is None:
                return f'UNBOUNDED {self.name}'
            else:
                return f'{value} {self.name}'


class GroupBy:
    """
    Class to build GROUP BY expressions. This is the basic building block to create more complex
    expressions. This class represents a grouping/partitioning by columns/series or expressions
    thereof.

    For more complex grouping expressions, this class should be extended.

    Instances of this class and subclasses are **immutable**. Any modifying operations should
    return a fresh copy.
    """
    def __init__(self, group_by_columns: List[Series]):
        from bach.series import SeriesAbstractMultiLevel
        self._index = {}

        for col in group_by_columns:
            if not isinstance(col, Series):
                raise ValueError(f'Unsupported argument type: {type(col)}')

            if (
                isinstance(col, SeriesAbstractMultiLevel)
                and any(lvl.expression.is_constant for lvl in col.levels.values())
            ):
                raise ValueError(
                    'Level in multi-level series has a constant expression, '
                    'please materialize first.'
                )
            if col.expression.is_constant:
                # We don't support this currently. If we allow this we would generate sql of the form (this
                # assumes the constants is '5'): `... group by (5)`. The sql is valid, it groups by the
                # fifth column of the query. But it doesn't make any sense to support that, the user should
                # be explicit in naming the columns that he wants to group on.
                raise ValueError('Grouping on a Series whose expression is a constant, is not supported.')
            if col.expression.has_windowed_aggregate_function:
                raise ValueError('Window functions can not be used to group, '
                                 'please materialize first.')
            if col.expression.has_aggregate_function:
                raise ValueError('Aggregate functions can not be used to group, '
                                 'please materialize first.')

            # index columns have no index themselves, and can also be evaluated without group_by as
            # they will not be aggregated by this group_by
            self._index[col.name] = col.copy_override(index={}, group_by=None, order_by=[])

    @property
    def index(self) -> Dict[str, Series]:
        return copy(self._index)

    def __eq__(self, other):
        if not isinstance(other, GroupBy):
            return False
        return (
            list(self._index.keys()) == list(other.index.keys()) and
            all(self._index[n].equals(other.index[n], recursion='GroupBy')
                for n in self._index.keys())
        )

    def get_index_expressions(self) -> List[Expression]:
        from bach.series import SeriesAbstractMultiLevel
        exprs = []
        for pos, idx in enumerate(self.index.values()):
            if isinstance(idx, SeriesAbstractMultiLevel):
                exprs += idx.level_expressions
                continue

            # BigQuery MUST use a reference for considering expression in GROUP BY clause
            # for example:
            # a) SELECT `x` + 1 as `x`, sum(y) from table GROUP BY `x`
            # is not equivalent to:
            # b) SELECT `x` + 1 as `x`, sum(y) from table GROUP BY `x` + 1
            # GROUP BY expression from b) actually is (`x` + 1) + 1
            # meanwhile GROUP BY expression from a) is (`x` + 1)
            if is_bigquery(idx.engine):
                exprs.append(Expression.column_reference(idx.name))

            # Athena allows grouping by expressions, column names and ordinals. A problem arises
            # when the series' expression is way too complex to handle (expression dependency)
            # In order to avoid this type of situations, is best to use ordinals, as we can make reference
            # to complex expressions by just using there ordinal position in the SELECT clause.
            # By default, index columns are the first in the SELECT clause. So, we can trust the order of
            # the GroupBy.index.
            # https://prestodb.io/docs/current/sql/select.html#group-by-clause
            elif is_athena(idx.engine) and idx.expression != Expression.column_reference(idx.name):
                exprs.append(ConstValueExpression.construct(f'{pos + 1}'))

            # this logic does not applies in Postgres
            else:
                exprs.append(idx.expression)

        return exprs

    def get_group_by_column_expression(self) -> Optional[Expression]:
        """
        Get the group_by expression, including all the relevant columns and the way of grouping,
        but without the "group by" clause, as these potentially have to be nested into one group-by.

        Will return None if self.index is empty, i.e. there was a group-by but no columns are
        specified.
        """
        if not self.index:
            return None

        exprs = self.get_index_expressions()
        fmt_strs = ['{}'] * len(exprs)

        fmtstr = ', '.join(fmt_strs)
        return Expression.construct(f'{fmtstr}', *exprs)

    def copy_override_base_node(self: G, base_node: BachSqlModel) -> G:
        new_cols = [col.copy_override(base_node=base_node) for col in self.index.values()]
        return self.__class__(group_by_columns=new_cols)


class Cube(GroupBy):
    """
    Very simple abstraction to support cubes
    """
    def __init__(self, group_by_columns: List[Series]):
        if len(group_by_columns) == 0:
            raise ValueError('Can not create a cube without group by columns')
        super().__init__(group_by_columns)

    def get_group_by_column_expression(self) -> Optional[Expression]:
        # help mypy, our parent always returns an Expression
        return Expression.construct('cube ({})', cast(Expression, super().get_group_by_column_expression()))


class Rollup(GroupBy):
    """
    Very simple abstraction to support rollups
    """

    def __init__(self, group_by_columns: List[Series]):
        if len(group_by_columns) == 0:
            raise ValueError('Can not create a rollup without group by columns')
        super().__init__(group_by_columns)

    def get_group_by_column_expression(self) -> Optional[Expression]:
        # help mypy, our parent always returns an Expression
        return Expression.construct('rollup ({})', cast(Expression, super().get_group_by_column_expression()))


class GroupingList(GroupBy):
    """
    Abstraction to support SQL's
    GROUP BY (colA,colB), CUBE(ColC,ColD), ROLLUP(ColC,ColE)
    like expressions
    """
    _grouping_list: List[GroupBy]

    def __init__(self, grouping_list: List[GroupBy]):
        """
        Given the list of groupbys, construct a combined groupby
        """
        self._grouping_list = grouping_list

        group_by_columns = {}

        for g in grouping_list:
            if not isinstance(g, GroupBy):
                raise ValueError("Only GroupBy items are supported")

            for name, series in g._index.items():
                if name not in group_by_columns:
                    group_by_columns[name] = series

        super().__init__(group_by_columns=list(group_by_columns.values()))

    def get_group_by_column_expression(self) -> Optional[Expression]:
        grouping_optional_expr_list = [g.get_group_by_column_expression() for g in self._grouping_list]
        grouping_expr_list: List[Expression] = []
        for expr in grouping_optional_expr_list:
            if expr is None:
                grouping_expr_list.append(Expression.construct(''))
            else:
                grouping_expr_list.append(expr)

        fmtstr = ', '.join(["({})"] * len(grouping_expr_list))
        return Expression.construct(fmtstr, *grouping_expr_list)


class GroupingSet(GroupingList):
    """
    Abstraction to support SQLs
    GROUP BY GROUPING SETS ((colA,colB),(ColA),(ColC))
    """
    def get_group_by_column_expression(self) -> Optional[Expression]:
        grouping_optional_expr_list = [g.get_group_by_column_expression() for g in self._grouping_list]
        grouping_expr_list: List[Expression] = []
        for expr in grouping_optional_expr_list:
            if expr is None:
                grouping_expr_list.append(Expression.construct(''))
            else:
                grouping_expr_list.append(expr)

        fmtstr = ', '.join(["({})"] * len(grouping_expr_list))
        fmtstr = f'grouping sets ({fmtstr})'
        return Expression.construct(fmtstr, *grouping_expr_list)


class Window(GroupBy):
    """
    Class representing an "immutable" window as defined in the SQL standard. Any operation on this
    class that would alter it returns a fresh copy. It can be reused many time if the window
    is reused.

    A Window for us is basically a partitioned, sorted view on a DataFrame, where the frame
    boundaries as given in the constructor, or in set_frame_clause(), define the window.

    A frame is defined in PG as follows:
    (See https://www.postgresql.org/docs/14/sql-expressions.html#SYNTAX-WINDOW-FUNCTIONS)

    { RANGE | ROWS } frame_start
    { RANGE | ROWS } BETWEEN frame_start AND frame_end
    where frame_start and frame_end can be one of

    UNBOUNDED PRECEDING
    value PRECEDING
    CURRENT ROW
    value FOLLOWING
    UNBOUNDED FOLLOWING

    The frame_clause specifies the set of rows constituting the window frame, for those window
    functions that act on the frame instead of the whole partition.

    If frame_end is omitted it
    defaults to CURRENT ROW.

    Restrictions are that
    - frame_start cannot be UNBOUNDED FOLLOWING,
    - frame_end cannot be UNBOUNDED PRECEDING
    - frame_end choice cannot appear earlier in the above list than the frame_start choice:
        for example RANGE BETWEEN CURRENT ROW AND value PRECEDING is not allowed.

    The default framing option is RANGE UNBOUNDED PRECEDING, which is the same as
    RANGE BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW; it sets the frame to be all rows from
    the partition start up through the current row's last peer in the ORDER BY ordering
    (which means all rows if there is no ORDER BY).

    In general, UNBOUNDED PRECEDING means that the frame starts with the first row of the
    partition, and similarly UNBOUNDED FOLLOWING means that the frame ends with the last row
    of the partition (regardless of RANGE or ROWS mode).

    In ROWS mode, CURRENT ROW means that the frame starts or ends with the current row;
    In RANGE mode it means that the frame starts or ends with the current row's first or
    last peer in the ORDER BY ordering.

    The value PRECEDING and value FOLLOWING cases are currently only allowed in ROWS mode.
    They indicate that the frame starts or ends with the row that many rows before or after
    the current row.

    value must be an integer expression not containing any variables, aggregate functions,
    or window functions. The value must not be null or negative; but it can be zero,
    which selects the current row itself.
    """
    def __init__(self,
                 dialect: Dialect,
                 group_by_columns: List['Series'],
                 order_by: List[SortColumn],
                 na_position: str = 'last',
                 mode: WindowFrameMode = WindowFrameMode.RANGE,
                 start_boundary: Optional[WindowFrameBoundary] = WindowFrameBoundary.PRECEDING,
                 start_value: int = None,
                 end_boundary: Optional[WindowFrameBoundary] = WindowFrameBoundary.CURRENT_ROW,
                 end_value: int = None,
                 min_values: int = None):
        """
        Define a window on a DataFrame, by giving the partitioning series and the frame definition
        :see: class definition for more info on the frame definition, and GroupBy for more
              info on grouping / partitioning
        :param na_position: Either 'first' or 'last'. When ordering values, put null values before all other
            values if 'first', and put null values after all other values if 'last'.
        """
        super().__init__(group_by_columns=group_by_columns)

        if mode is None:
            raise ValueError("Mode needs to be defined")

        if na_position not in ['first', 'last']:
            raise ValueError(f'"{na_position}" is not a valid value for `na_position` param.')

        if start_boundary is None and end_boundary is not None:
            raise ValueError("start_boundary needs to be defined if end_boundary is present.")

        if start_boundary == WindowFrameBoundary.FOLLOWING and start_value is None:
            raise ValueError("Start of frame can not be unbounded following")

        if end_boundary == WindowFrameBoundary.PRECEDING and end_value is None:
            raise ValueError("End of frame can not be unbounded preceding")

        if (start_value is not None and start_value < 0) or \
                (end_value is not None and end_value < 0):
            raise ValueError("start_value and end_value must be greater than or equal to zero.")

        if mode == WindowFrameMode.RANGE \
                and (start_value is not None or end_value is not None):
            raise ValueError("start_value or end_value cases only supported in ROWS mode.")

        if start_boundary is not None and end_boundary is not None:
            if start_boundary.value > end_boundary.value:
                raise ValueError("frame boundaries defined in wrong order.")

            if start_boundary == end_boundary:
                if start_boundary == WindowFrameBoundary.PRECEDING \
                        and start_value is not None \
                        and end_value is not None \
                        and start_value < end_value:
                    raise ValueError("frame boundaries defined in wrong order.")

            if start_boundary == end_boundary:
                if start_boundary == WindowFrameBoundary.FOLLOWING \
                        and start_value is not None \
                        and end_value is not None \
                        and start_value > end_value:
                    raise ValueError("frame boundaries defined in wrong order.")

        self._dialect = dialect
        self._mode = mode
        self._start_boundary = start_boundary
        self._start_value = start_value
        self._end_boundary = end_boundary
        self._end_value = end_value
        self._min_values = 0 if min_values is None else min_values

        # TODO This should probably be an expression
        self._frame_clause: str = ''
        if start_boundary and end_boundary is None:
            self._frame_clause = f'{mode.name} {start_boundary.frame_clause(start_value)}'
        elif start_boundary and end_boundary:
            self._frame_clause = f'{mode.name} BETWEEN {start_boundary.frame_clause(start_value)}'\
                                f' AND {end_boundary.frame_clause(end_value)}'

        self._order_by = order_by
        self._na_position = na_position

    @property
    def frame_clause(self) -> str:
        return self._frame_clause

    @property
    def order_by(self) -> List[SortColumn]:
        return self._order_by

    @property
    def min_values(self) -> int:
        return self._min_values

    def set_frame_clause(self,
                         mode:
                         WindowFrameMode = WindowFrameMode.RANGE,
                         start_boundary:
                         Optional[WindowFrameBoundary] = WindowFrameBoundary.PRECEDING,
                         start_value: int = None,
                         end_boundary:
                         Optional[WindowFrameBoundary] = WindowFrameBoundary.CURRENT_ROW,
                         end_value: int = None) -> 'Window':
        """
        Convenience function to clone this window with new frame parameters
        :see: __init__()
        """
        return Window(
            dialect=self._dialect,
            group_by_columns=list(self._index.values()),
            order_by=self._order_by,
            na_position=self._na_position,
            mode=mode,
            start_boundary=start_boundary, start_value=start_value,
            end_boundary=end_boundary, end_value=end_value,
        )

    def get_index_expressions(self) -> List[Expression]:
        from bach.series import SeriesAbstractMultiLevel
        exprs = []
        for idx in self.index.values():
            if isinstance(idx, SeriesAbstractMultiLevel):
                exprs += idx.level_expressions
                continue

            exprs.append(idx.expression)

        return exprs

    def get_window_expression(self, window_func: Expression) -> Expression:
        """
        Given the window_func generate a statement like:
            {window_func} OVER (PARTITION BY .. ORDER BY ... frame_clause)
        """
        # TODO implement NULLS FIRST / NULLS LAST, probably not here but in the sorting logic.
        order_by = get_order_by_expression(
            dialect=self._dialect, order_by=self._order_by, na_position=self._na_position,
        )

        if self.frame_clause is None:
            frame_clause = ''
        else:
            frame_clause = self.frame_clause

        index_exprs = []
        if len(self._index) == 0:
            partition_fmt = ''
        else:
            index_exprs = self.get_index_expressions()
            fmt_stmts = ['{}'] * len(index_exprs)
            partition_fmt = 'partition by ' + ', '.join(fmt_stmts)

        over_fmt = f'over ({partition_fmt} {{}} {frame_clause})'
        over_expr = Expression.construct(over_fmt, *index_exprs, order_by)

        if self._min_values is None or self._min_values == 0:
            return WindowFunctionExpression.construct(f'{{}} {{}}', window_func, over_expr)
        else:
            # Only return a value when then minimum amount of observations (including NULLs)
            # has been reached.
            return WindowFunctionExpression.construct(f"""
                case when (count(1) {{}}) >= {self._min_values}
                then {{}} {{}}
                else NULL end""", over_expr, window_func, over_expr)

    def get_group_by_column_expression(self) -> Optional[Expression]:
        """
        On a Window, there is no default group_by clause
        """
        return None


def get_order_by_expression(
    dialect: Dialect, order_by: List['SortColumn'], na_position: str = 'last',
) -> Expression:
    """
    INTERNAL: Convert order_by into an order by expression that is usable inside an aggregation or window
    function.

    Note: The default ordering mimics the default behaviour of pandas, where nulls are always sorted last.

    :param order_by: List of SortColumns defining the ordering
    :param na_position: By default (=`last`) the sorting is 'asc null last' and 'desc nulls last'. If
        `na_position` is `first`, then the sorting is set to 'asc nulls first', 'desc nulls first'
    :return: An Expression containing a complete order by clause of the form 'order by ...'. Will return
        an empty Expression if the specified `order_by` list is empty
    """

    # This logic is partially duplicating DataFrame._get_order_by_clause(), but we can't quite re-use
    # that as some of the checks there are not applicable, e.g. it's fine to use a column that we are not
    # grouping on in an aggregate order by.
    if not order_by:
        return Expression.construct('')

    if na_position not in ['last', 'first']:
        raise ValueError(
            f'"{na_position}" is not a valid value for `na_position` param.'
        )

    asc_expr = Expression.raw('asc')
    desc_expr = Expression.raw('desc')
    if na_position == 'last':
        nulls_expr = Expression.construct('nulls last')
    else:
        nulls_expr = Expression.construct('nulls first')

    expressions: List[Expression] = []
    for sc in order_by:
        if sc.expression.has_multi_level_expressions:
            multi_lvl_exprs = [
                level_expr for level_expr in sc.expression.data if isinstance(level_expr, Expression)
            ]
        else:
            multi_lvl_exprs = [sc.expression]
        for level_expr in multi_lvl_exprs:
            expr = Expression.construct('{} {}', level_expr,  asc_expr if sc.asc else desc_expr)
            if is_bigquery(dialect):
                # Big Query does not support NULLS LAST and NULLS FIRST (might generate some errors)
                # in window order by, therefore we require to simulate it.
                simulate_nulls_last_expr = Expression.construct(
                    '({} is null) {}', level_expr, asc_expr if na_position == 'last' else desc_expr
                )
                expressions.append(simulate_nulls_last_expr)
            else:
                expr = Expression.construct('{} {}', expr, nulls_expr)

            expressions.append(expr)

    return Expression.construct('order by {}', join_expressions(expressions))



"""
Copyright 2021 Objectiv B.V.
"""
from typing import Tuple, Dict

import numpy
import pandas
from sqlalchemy.engine import Engine, Dialect

from bach import DataFrame, get_series_type_from_dtype
from bach.types import value_to_dtype, DtypeOrAlias, Dtype
from bach.expression import Expression, join_expressions
from bach.utils import get_sql_column_name, get_name_to_column_mapping
from sql_models.model import CustomSqlModelBuilder
from sql_models.util import quote_identifier, DatabaseNotSupportedException, is_postgres, is_bigquery, \
    is_athena


def from_pandas(engine: Engine,
                df: pandas.DataFrame,
                convert_objects: bool,
                name: str,
                materialization: str,
                if_exists: str = 'fail') -> DataFrame:
    """
    See DataFrame.from_pandas() for docstring.
    """
    if materialization == 'cte':
        return from_pandas_ephemeral(engine=engine, df=df, convert_objects=convert_objects, name=name)
    if materialization == 'table':
        return from_pandas_store_table(
            engine=engine,
            df=df,
            convert_objects=convert_objects,
            table_name=name,
            if_exists=if_exists
        )
    raise ValueError(f'Materialization should either be "cte" or "table", value: {materialization}')


def from_pandas_store_table(engine: Engine,
                            df: pandas.DataFrame,
                            convert_objects: bool,
                            table_name: str,
                            if_exists: str = 'fail') -> DataFrame:
    """
    Instantiate a new DataFrame based on the content of a Pandas DataFrame. This will first write the
    data to a database table using pandas' df.to_sql() method.
    Supported dtypes are 'int64', 'float64', 'string', 'datetime64[ns]', 'bool'


    :param engine: db connection
    :param df: Pandas DataFrame to instantiate as DataFrame
    :param convert_objects: If True, columns of type 'object' are converted to 'string' using the
        pd.convert_dtypes() method where possible.
    :param table_name: name of the sql table the Pandas DataFrame will be written to
    :param if_exists: {'fail', 'replace', 'append'}, default 'fail'
        How to behave if the table already exists:
        * fail: Raise a ValueError.
        * replace: Drop the table before inserting new values.
        * append: Insert new values to the existing table.
    """
    # todo add dtypes argument that explicitly let's you set the supported dtypes for pandas columns
    df_copy, index_dtypes, all_dtypes, name_to_column_mapping = _from_pd_shared(
        dialect=engine.dialect,
        df=df,
        convert_objects=convert_objects,
        cte=False
    )

    df_copy = df_copy.rename(columns=name_to_column_mapping)

    if is_athena(engine):
        _athena_pandas_to_sql(
            engine=engine,
            df=df_copy,
            table_name=table_name,
            if_exists=if_exists,
        )
    else:
        conn = engine.connect()
        df_copy.to_sql(name=table_name, con=conn, if_exists=if_exists, index=False)
        conn.close()

    index = list(index_dtypes.keys())
    return DataFrame.from_table(
        engine=engine,
        table_name=table_name,
        index=index,
        all_dtypes=all_dtypes,
        name_to_column_mapping=name_to_column_mapping
    )


def _athena_pandas_to_sql(df: pandas.DataFrame, table_name: str, engine: Engine, if_exists: str) -> None:
    """
    Helper function for writing records from a pandas DataFrame to Amazon Athena.
    Currently, pandas.DataFrame.to_sql is broken for Athena, as an alternative
    we use `pyathena.panda.util.to_sql`, which stores the records into the specified `location` parameter
    in the engine's url. In case the location is not specified, s3_staging_dir parameter will be used instead.
    """
    from pyathena.pandas.util import to_sql
    from pyathena.sqlalchemy_athena import AthenaDialect
    from pyathena.connection import Connection

    connection_args = AthenaDialect().create_connect_args(engine.url)[1]

    # The location of the Amazon S3 table is specified by the location parameter in the connection string.
    # If location is not specified, s3_staging_dir parameter will be used
    # s3://{location or s3_staging_dir}/{schema}/{table}/
    location = connection_args.get('location') or connection_args["s3_staging_dir"]
    location += "/" if not location.endswith("/") else ""
    location += f'{table_name}/'

    # Pandas.to_sql is not used as pyAthena's SQLAlchemy DDL Compiler replaces BIGINT columns
    # with INT db type, most likely a bug from pyathena as it raises a ParseException when trying
    # to execute the CREATE statement.
    # Fortunately, pyathena's to_sql function creates the correct DDL statement but requires a
    # pyathena Connection instead

    # For more information: https://pypi.org/project/pyathena/#pandas
    with Connection(**connection_args) as conn:
        to_sql(
            df,
            name=table_name,
            schema=connection_args['schema_name'],
            conn=conn,
            if_exists=if_exists,
            index=False,
            location=location
        )


def from_pandas_ephemeral(
        engine: Engine,
        df: pandas.DataFrame,
        convert_objects: bool,
        name: str
) -> DataFrame:
    """
    Instantiate a new DataFrame based on the content of a Pandas DataFrame. The data will be represented
    using a `select * from values()` query, or something similar depending on the database dialect.

    Warning: This method is only suited for small quantities of data.
    For anything over a dozen kilobytes of data it is recommended to store the data in a table in
    the database, e.g. by using the from_pd_store_table() function.

    Supported dtypes are 'int64', 'float64', 'string', 'datetime64[ns]', 'bool'

    :param engine: db connection
    :param df: Pandas DataFrame to instantiate as DataFrame
    :param convert_objects: If True, columns of type 'object' are converted to 'string' using the
        pd.convert_dtypes() method where possible.
    """
    # todo add dtypes argument that explicitly let's you set the supported dtypes for pandas columns
    df_copy, index_dtypes, all_dtypes, name_to_column_mapping = _from_pd_shared(
        dialect=engine.dialect,
        df=df,
        convert_objects=convert_objects,
        cte=True
    )

    column_series_type = [get_series_type_from_dtype(dtype) for dtype in all_dtypes.values()]

    per_row_expr = []
    for row in df_copy.itertuples():
        per_column_expr = []
        # Access the columns in `row` by index rather than by name. Because if a name starts with an
        # underscore (e.g. _index_skating_order) it will not be available as attribute.
        # so we use `row[i]` instead of getattr(row, column_name).
        # start=1 is to account for the automatic index that pandas adds
        for i, series_type in enumerate(column_series_type, start=1):
            val = row[i]
            per_column_expr.append(
                series_type.value_to_expression(dialect=engine.dialect, value=val, dtype=series_type.dtype)
            )
        row_expr = Expression.construct('({})', join_expressions(per_column_expr))
        per_row_expr.append(row_expr)
    all_values_str = join_expressions(per_row_expr, join_str=',\n').to_sql(engine.dialect)

    dialect = engine.dialect
    if is_postgres(engine) or is_athena(engine):
        # We are building sql of the form:
        #     select * from (values
        #         ('row 1', cast(1234 as bigint), cast(-13.37 as double precision)),
        #         ('row 2', cast(1337 as bigint), cast(31.337 as double precision))
        #     ) as t("a", "b", "c")
        column_names_expressions = []
        for series_name in all_dtypes.keys():
            sql_column_name = name_to_column_mapping.get(series_name, series_name)
            column_names_expressions.append(Expression.raw(quote_identifier(dialect, sql_column_name)))
        column_names_expr = join_expressions(column_names_expressions)
        column_names_str = column_names_expr.to_sql(engine.dialect)
        sql = f'select * from (values \n{all_values_str}\n) as t({column_names_str})\n'
    elif is_bigquery(engine):
        # We are building sql of the form:
        #     select * from UNNEST([
        #         STRUCT<`a` STRING, `b` INT64, `c` FLOAT64>
        #         ('row 1', 1234, cast(-13.37 as FLOAT64))
        #         ('row 2', 1337, cast(31.337 as FLOAT64))
        #     ])
        sql_column_name_types = []
        for series_name, dtype in all_dtypes.items():
            sql_column_name = name_to_column_mapping.get(series_name, series_name)
            db_col_name = quote_identifier(dialect=engine.dialect, name=sql_column_name)
            db_dtype = get_series_type_from_dtype(dtype).get_db_dtype(dialect=engine.dialect)
            sql_column_name_types.append(f'{db_col_name} {db_dtype}')
        sql_struct = f'STRUCT<{", ".join(sql_column_name_types)}>'
        sql = f'select * from UNNEST([{sql_struct} \n{all_values_str}\n])\n'
    else:
        raise DatabaseNotSupportedException(engine)

    model_builder = CustomSqlModelBuilder(sql=sql, name=name)
    sql_model = model_builder()

    index = list(index_dtypes.keys())
    return DataFrame.from_model(
        engine=engine,
        model=sql_model,
        index=index,
        all_dtypes=all_dtypes,
        name_to_column_mapping=name_to_column_mapping
    )


def _assert_column_names_valid(dialect: Dialect, df: pandas.DataFrame):
    """
    Performs three checks on the columns (not on the indices) of the DataFrame:
    1) All Series must have a string as name
    2) No duplicated Series names
    3) All Series names either are a valid column name or can be mapped to a valid column name in the given
        sql dialect.
    Will raise a ValueError if any of the checks fails
    """
    names = list(df.columns)
    not_strings = [name for name in names if not isinstance(name, str)]
    if not_strings:
        raise ValueError(f'Not all columns names are strings. Non-string names: {not_strings}')
    if len(set(names)) != len(names):
        raise ValueError(f'Duplicate column names in: {names}')

    for name in names:
        # Will raise an exception if this cannot be turned into a legal column_name
        get_sql_column_name(dialect=dialect, name=name)


def _from_pd_shared(
        dialect: Dialect,
        df: pandas.DataFrame,
        convert_objects: bool,
        cte: bool
) -> Tuple[pandas.DataFrame, Dict[str, Dtype], Dict[str, Dtype], Dict[str, str]]:
    """
    Pre-processes the given Pandas DataFrame, and do some checks. This results in a new DataFrame and three
    dictionaries with meta data. The returned DataFrame contains all processed data in the data columns, this
    includes any columns that were originally indices. The index_dtypes dict will indicate which of the data
    columns should be indexes. The name_to_column_mapping dict defines what column name should be used for
    each series; depending on the dialect certain names are not allowed as column names so we cannot just use
    the series names.

    Pre-processing/checks:
    1)  Add index if missing
    2a) Convert string columns to string dtype (if convert_objects)
    2b) Set content of columns of dtype other than `supported_pandas_dtypes` to supported types
        (if convert_objects & cte)
    3)  Check that the dtypes are supported
    4)  extract index_dtypes and dtypes dictionaries
    5) Check: all column names (after previous steps) are set, unique, and can be mapped to valid column
        names for the given dialect.

    :return: Tuple:
        * Modified copy of Pandas DataFrame. No index columns, the original index columns are regular
            columns in this copy.
        * index_dtypes dict. Mapping of index names to the dtype.
        * all_dtypes dict. Mapping of all series names to dtypes, includes data and index columns.
        * name_to_column_mapping: mapping of series name, to sql column name
    """
    index = []

    for idx, name in enumerate(df.index.names):
        if name is None:
            name = f'_index_{idx}'
        else:
            name = f'_index_{name}'

        index.append(name)

    df_copy = df.copy()
    df_copy.index.set_names(index, inplace=True)
    # set the index as normal columns, this makes it easier to convert the dtype
    df_copy.reset_index(inplace=True)

    supported_pandas_dtypes = ['int64', 'float64', 'string', 'datetime64[ns]', 'bool', 'int32']
    all_dtypes_or_alias: Dict[str, DtypeOrAlias] = {}
    for series_name in df_copy.columns:
        series_name = str(series_name)
        dtype = df_copy[series_name].dtype.name

        if dtype in supported_pandas_dtypes:
            all_dtypes_or_alias[series_name] = dtype
            continue

        if df_copy[series_name].dropna().empty:
            raise ValueError(f'{series_name} column has no non-nullable values, cannot infer type.')

        if convert_objects:
            df_copy[series_name] = df_copy[series_name].convert_dtypes(
                convert_integer=False, convert_boolean=False, convert_floating=False,
            )
            dtype = df_copy[series_name].dtype.name

        if dtype not in supported_pandas_dtypes and not(cte and convert_objects):
            raise TypeError(f'unsupported dtype for {series_name}: {dtype}')

        if cte and convert_objects:
            non_nullables = df_copy[series_name].dropna()
            types = non_nullables.apply(type).unique()
            if len(types) != 1:
                raise TypeError(f'multiple types found in column {series_name}: {types}')
            dtype = value_to_dtype(non_nullables.iloc[0])

        all_dtypes_or_alias[series_name] = dtype

    _assert_column_names_valid(dialect=dialect, df=df_copy)

    all_dtypes = {name: get_series_type_from_dtype(dtype_or_alias).dtype
                  for name, dtype_or_alias in all_dtypes_or_alias.items()}
    index_dtypes = {index_name: all_dtypes[index_name] for index_name in index}

    name_to_column_mapping = get_name_to_column_mapping(dialect=dialect, names=all_dtypes.keys())

    df_copy = df_copy.replace({numpy.nan: None})
    return df_copy, index_dtypes, all_dtypes, name_to_column_mapping



"""
Copyright 2021 Objectiv B.V.
"""
import itertools
from copy import copy
from enum import Enum
from typing import Union, List, Tuple, Optional, Dict, Set, Hashable, cast, NamedTuple, Sequence, Mapping

from sqlalchemy.engine import Dialect

from bach import DataFrameOrSeries, DataFrame, ColumnNames, Series, SeriesBoolean
from bach.dataframe import DtypeNamePair

from bach.expression import (
    Expression, join_expressions, TableColumnReferenceToken,  ExpressionToken, ColumnReferenceToken,
)
from bach.utils import ResultSeries, get_result_series_dtype_mapping, get_sql_column_name, \
    get_name_to_column_mapping
from sql_models.constants import NotSet, not_set
from sql_models.model import Materialization, CustomSqlModelBuilder, SqlModel, SqlModelSpec
from bach.sql_model import BachSqlModel, construct_references


class How(Enum):
    """ Enum with all valid values of 'how' parameter """
    left = 'left'
    right = 'right'
    outer = 'outer'
    inner = 'inner'
    cross = 'cross'


_LEFT_NODE_ALIAS = 'l'
_RIGHT_NODE_ALIAS = 'r'


class MergeOn(NamedTuple):
    left: List[str]
    right: List[str]
    conditional: List[SeriesBoolean]

    @property
    def is_empty(self) -> bool:
        return not (self.left or self.right or self.conditional)


def _verify_on_conflicts(
    left: DataFrame,
    right: DataFrameOrSeries,
    how: How,
    on: Optional[Sequence[Union[str, SeriesBoolean]]],
    left_on: Optional[ColumnNames],
    right_on: Optional[ColumnNames],
    left_index: bool,
    right_index: bool,
) -> None:
    """
    Verifies that provided on parameters are valid for merge operation.
    Rules for valid merge:
        1. Any of "on", "left_on", "left_index", "right_on", "right_index"
            should be provided if how != 'cross'
        2. Both "left_on" and "left_index" must not be specified at the same time.
        3. Both "right_on" and "right_index" must not be specified at the same time.
        4. If "left_on"/"left_index" is specified, "right_on"/"right_index" must be specified as well.
        5. If "on" is specified (ignoring SeriesBoolean):
            "left_on", "left_index", "right_on", "right_index" must be None
        6. If "on" contains SeriesBoolean, each series should make reference to both left and right objects to
           be merged.
    """
    if how == How.cross and (on or left_on or right_on or left_index or right_index):
        raise ValueError(
            'Cannot specify on, left_on, right_on, left_index, or right_index if how == "cross"'
        )

    if (left_on is not None) and left_index:
        raise ValueError('Cannot specify both left_on and left_index.')

    if (right_on is not None) and right_index:
        raise ValueError('Cannot specify both right_on and right_index.')

    if bool(left_on or left_index) ^ bool(right_on or right_index):
        raise ValueError(
            'Either both left_on/left_index and right_on/right_index should be specified, '
            'or both should be None.'
        )

    on_conditions = [o for o in on or [] if isinstance(o, SeriesBoolean)] if on else []
    on_columns = [o for o in on or [] if not isinstance(o, SeriesBoolean)] if on else None

    if on_columns and (left_on or left_index) and (right_on or right_index):
        raise ValueError('Either specify on or, left_on and right_on, but not all three')

    if not on_conditions:
        return None

    if on_conditions and left.base_node == right.base_node:
        raise ValueError('"on" based SeriesBooleans is valid only when left.base_node != right.base_node. ')

    for col in on_conditions:
        refs = col.base_node.references.values()
        if not (left.base_node in refs and right.base_node in refs):
            raise ValueError('boolean series must have both base_nodes to be merged as references.')


def _determine_merge_on(
    left: DataFrame,
    right: DataFrameOrSeries,
    how: How,
    on: Optional[Sequence[Union[str, SeriesBoolean]]],
    left_on: Optional[ColumnNames],
    right_on: Optional[ColumnNames],
    left_index: bool,
    right_index: bool,
) -> MergeOn:
    """
    Determine the columns that should be equal for the merge. Both for the left and the right
    dataframe/series a list of strings is returned indicating the names of the columns that should be
    matched.
    :return: Tuple containing left, right and conditional on
    """
    _verify_on_conflicts(
        left=left,
        right=right,
        how=how,
        on=on,
        left_on=left_on,
        right_on=right_on,
        left_index=left_index,
        right_index=right_index,
    )
    if how == How.cross:
        return MergeOn(left=[], right=[], conditional=[])

    left_on = left_on or list(_get_index_names(left)) if left_on or left_index else None
    right_on = right_on or list(_get_index_names(right)) if right_on or right_index else None

    if on is not None:
        final_on = [o for o in on or [] if not isinstance(o, SeriesBoolean)]
    else:
        final_on = list(_get_data_columns(left) & _get_data_columns(right))

    final_left_on = _get_x_on(final_on, left_on, 'left_on')
    final_right_on = _get_x_on(final_on, right_on, 'right_on')

    if len(final_left_on) != len(final_right_on):
        raise ValueError(
            f'Len of left_on ({final_left_on}) does not match that of right_on ({final_right_on}).')

    on_conditions = [o for o in on or [] if isinstance(o, SeriesBoolean)] if on else []

    if not final_left_on and not on_conditions:
        raise ValueError('No columns to perform merge on')

    missing_left = set(final_left_on) - _get_all_series_names(left)
    missing_right = set(final_right_on) - _get_all_series_names(right)

    if missing_left:
        raise ValueError(f'Specified column(s) do not exist. left_on: {left_on}. missing: {missing_left}')
    if missing_right:
        raise ValueError(f'Specified column(s) do not exist. right_on: {right_on}. missing: {missing_right}')

    return MergeOn(left=final_left_on, right=final_right_on, conditional=on_conditions)


def _get_data_columns(df_series: DataFrameOrSeries) -> Set[str]:
    """ Get set with the names of all data columns. Works for both dataframe and series. """
    if isinstance(df_series, DataFrame):
        return set(df_series.data_columns)
    if isinstance(df_series, Series):
        return {df_series.name}
    raise TypeError(f'Expected bach.DataFrame or bach.Series, got {type(df_series)}')


def _get_index_names(df_series: DataFrameOrSeries) -> Set[str]:
    """ Get set the names of the index columns. Works for both dataframe and series. """
    if df_series.index:
        return set(df_series.index.keys())
    else:
        return set()


def _get_all_series_names(df_series: DataFrameOrSeries) -> Set[str]:
    """ Get set with the names of all series. Works for both dataframe and series. """
    return _get_index_names(df_series) | _get_data_columns(df_series)


def _get_x_on(on: ColumnNames, x_on: Optional[ColumnNames], var_name: str) -> List[str]:
    """ Helper for _determine_left_on_right_on: Give `x_on` as a List[str], or default to `on`. """
    if isinstance(x_on, str):
        return [x_on]
    if isinstance(x_on, list):
        return x_on
    if x_on is None:
        if isinstance(on, str):
            return [on]
        if isinstance(on, list):
            return on
        raise ValueError(f'Type of on is not supported. Type: {type(on)}')
    raise ValueError(f'Type of {var_name} is not supported. Type: {type(x_on)}')


def _determine_result_columns(
    dialect: Dialect,
    left: DataFrame,
    right: DataFrameOrSeries,
    merge_on: MergeOn,
    suffixes: Tuple[str, str],
) -> Tuple[List[ResultSeries], List[ResultSeries]]:
    """
    Determine which columns should be in the DataFrame after merging left and right, with the given
    left_on and right_on values.
    """
    if not isinstance(right, (DataFrame, Series)):
        raise TypeError(f'Right should be DataFrameOrSeries type: {type(right)}')

    left_df = left.copy()
    right_df = right.copy() if isinstance(right, DataFrame) else right.to_frame()

    conflicting_on = {l_on for l_on, r_on in zip(merge_on.left, merge_on.right) if l_on == r_on}
    conflicting = (
        (set(left_df.index) | set(left_df.data)) & (set(right_df.index) | set(right_df.data))
    )
    # don't add suffixes to conflicted on columns
    # need to consider values from both objects (important when how = How.outer)
    conflicting -= conflicting_on

    # left dataframe has priority over index and data columns
    # final shared index and data series are based on left dataframe structure
    right_index = {}
    right_data = {}
    for series_name, series in right_df.all_series.items():
        data_columns = left_df.data_columns if series_name in conflicting_on else right_df.data_columns
        if series_name in data_columns:
            right_data[series_name] = series
        else:
            right_index[series_name] = series

    new_index_list = _get_merged_result_series(
        dialect=dialect,
        left_series=left_df.index,
        right_series=right_index,
        suffixes=suffixes,
        conflicting_names=conflicting,
        conflicting_on=conflicting_on,
    )

    new_data_list = _get_merged_result_series(
        dialect=dialect,
        left_series=left_df.data,
        right_series=right_data,
        suffixes=suffixes,
        conflicting_names=conflicting,
        conflicting_on=conflicting_on,
    )

    _check_no_column_name_conflicts(new_index_list + new_data_list)
    return new_index_list, new_data_list


def _check_no_column_name_conflicts(result_columns: List[ResultSeries]):
    """ Helper of _determine_result_columns, checks that there are no duplicate names in the list.  """
    seen = set()
    for rc in result_columns:
        if rc.name in seen:
            raise ValueError(f'Names are not unique. Result contains {rc.name} multiple times')
        seen.add(rc.name)


def _get_merged_result_series(
    dialect: Dialect,
    left_series: Dict[str, Series],
    right_series: Dict[str, Series],
    conflicting_names: Set[str],
    conflicting_on: Set[str],
    suffixes: Tuple[str, str],
) -> List[ResultSeries]:
    """ Helper of _determine_result_columns. """
    new_column_results: List[ResultSeries] = []
    for suffix, source_series in zip(suffixes, (left_series, right_series)):
        table_alias = _LEFT_NODE_ALIAS if suffix == suffixes[0] else _RIGHT_NODE_ALIAS
        for series_name, series in source_series.items():
            new_name = series_name
            expr = series.expression.resolve_column_references(dialect, table_alias)

            if series_name in conflicting_on:
                if table_alias == _RIGHT_NODE_ALIAS:
                    continue
                r_expr = right_series[series_name].expression.resolve_column_references(
                    dialect, _RIGHT_NODE_ALIAS,
                )
                expr = Expression.construct(f'COALESCE({{}}, {{}})', expr, r_expr)
            elif series_name in conflicting_names:
                new_name = series_name + suffix

            new_column_results.append(
                ResultSeries(
                    name=new_name,
                    expression=expr,
                    dtype=series.dtype,
                )
            )
    return new_column_results


def merge(
    left: DataFrame,
    right: DataFrameOrSeries,
    how: str,
    on: Union[str, 'SeriesBoolean', List[Union[str, 'SeriesBoolean']], None],
    left_on: Union[str, List[str],  None],
    right_on: Union[str, List[str], None],
    left_index: bool,
    right_index: bool,
    suffixes: Tuple[str, str]
) -> DataFrame:
    """
    See :py:meth:`bach.DataFrame.merge` for more information.
    """
    if how not in ('left', 'right', 'outer', 'inner', 'cross'):
        raise ValueError(f"how must be one of ('left', 'right', 'outer', 'inner', 'cross'), value: {how}")

    if left.group_by:
        left = left.materialize(node_name='merge_left')

    if right.group_by:
        if isinstance(right, Series):
            right = right.to_frame()
        right = right.materialize(node_name='merge_right')

    dialect = left.engine.dialect

    real_how = How(how)
    merge_on = _determine_merge_on(
        left=left,
        right=right,
        how=real_how,
        on=[on] if on is not None and not isinstance(on, list) else on,
        left_on=left_on,
        right_on=right_on,
        left_index=left_index,
        right_index=right_index
    )

    new_index_list, new_data_list = _determine_result_columns(
        dialect=dialect,
        left=left,
        right=right,
        merge_on=merge_on,
        suffixes=suffixes,
    )

    if isinstance(right, Series):
        from bach.savepoints import Savepoints
        right_savepoints = Savepoints()
        right_variables = {}
    else:
        right_savepoints = right.savepoints
        right_variables = right.variables
    # copy right_variables, and then overwrite with left. This means that the left variables 'win' in case
    # where the same variable name/dtype exist in both left and right
    variables = copy(right_variables)
    variables.update(left.variables)

    new_column_list = new_index_list + new_data_list
    name_to_column_mapping = get_name_to_column_mapping(dialect, [rs.name for rs in new_column_list])

    model = _get_merge_sql_model(
        dialect=dialect,
        left=left,
        right=right,
        how=real_how,
        merge_on=merge_on,
        new_column_list=new_column_list,
        variables=variables
    )

    return left.copy_override(
        engine=left.engine,
        base_node=model,
        index_dtypes=get_result_series_dtype_mapping(new_index_list),
        series_dtypes=get_result_series_dtype_mapping(new_data_list),
        group_by=None,
        order_by=[],  # merging resets any sorting
        savepoints=left.savepoints.merge(right_savepoints),
        variables=variables,
        name_to_column_mapping=name_to_column_mapping
    )


def revert_merge(base: DataFrame) -> Tuple[DataFrame, DataFrame]:
    """
    Splits a merged dataframe into two new frames that have the same structure as the original frames
    that were combined. Base dataframe should not be aggregated since original indexes can be lost.
    """
    if not isinstance(base.base_node, MergeSqlModel):
        raise Exception('can only revert merge if DataFrame base_node is MergeSqlModel instance.')

    if base.group_by:
        raise Exception('cannot revert merge on aggregated frame.')

    left_node = cast(BachSqlModel, base.base_node.references['left_node'])
    right_node = cast(BachSqlModel, base.base_node.references['right_node'])

    left_index, right_index = _determine_series_per_source(
        base=base, series_to_unmerge=base.index_columns, left_index={}, right_index={},
    )

    left_series, right_series = _determine_series_per_source(
        base=base,
        series_to_unmerge=list(set(base.base_node.series_names) - set(base.index_columns)),
        left_index=left_index,
        right_index=right_index,
    )

    left = base.copy_override(base_node=left_node, series=left_series, index=left_index)
    right = base.copy_override(base_node=right_node, series=right_series, index=right_index)

    return left, right


def _determine_series_per_source(
    base: 'DataFrame',
    series_to_unmerge: List[str],
    left_index: Dict[str, 'Series'],
    right_index: Dict[str, 'Series'],
) -> Tuple[Dict[str, 'Series'], Dict[str, 'Series']]:
    """
    Determine which series are on each original dataframe based on the node references
    from current MergeSqlModel. When identifying each column original source, a new series
    will be created and replicate the original structure.

    .. note::
    Series without table column references are added to both final results.
    """
    dialect = base.engine.dialect
    left_series = {}
    right_series = {}

    left_node = cast(BachSqlModel, base.base_node.references['left_node'])
    right_node = cast(BachSqlModel, base.base_node.references['right_node'])

    expressions_to_parse: Dict[str, List[Expression]] = {}

    merge_node = cast(MergeSqlModel, base.base_node)
    merge_on = merge_node.merge_on

    coalesced_series = [
        left_on for left_on, right_on in zip(merge_on.left, merge_on.right) if left_on == right_on
    ]
    for series_name, original_expr in base.base_node.column_expressions.items():
        if series_name not in series_to_unmerge:
            continue

        if series_name not in coalesced_series:
            expressions_to_parse[series_name] = [original_expr]
            continue

        # removes first-level tokens from COALESCE expression
        # I.e COALESCE(expr1, expr2)
        # expression_to_parse[series] = [expr1, expr2]
        expressions_to_parse[series_name] = [
            sub_expr
            for sub_expr in original_expr.data
            if isinstance(sub_expr, Expression)
        ]

    default_series = base.all_series[base.data_columns[0]]
    for series_name, expressions in expressions_to_parse.items():
        if series_name in base.all_series:
            series = base.all_series[series_name]
        else:
            series = default_series.copy()

        for sub_expr in expressions:
            table_name, column_name, expr = sub_expr.remove_table_column_references()

            if table_name == _LEFT_NODE_ALIAS or not sub_expr.has_table_column_references:
                if column_name:
                    original_series_name = _get_series_name_from_column_name(
                        model=left_node,
                        dialect=dialect,
                        column_name=column_name
                    )
                else:
                    original_series_name = series_name

                left_series[original_series_name] = series.copy_override(
                    base_node=left_node,
                    index=left_index,
                    name=original_series_name,
                    expression=expr,
                )
            if table_name == _RIGHT_NODE_ALIAS or not sub_expr.has_table_column_references:
                if column_name:
                    original_series_name = _get_series_name_from_column_name(
                        model=right_node,
                        dialect=dialect,
                        column_name=column_name
                    )
                else:
                    original_series_name = series_name
                right_series[original_series_name] = series.copy_override(
                    base_node=right_node,
                    index=right_index,
                    name=original_series_name,
                    expression=expr,
                )

    ordered_left_series = _order_series_based_on_node(left_node, left_series)
    ordered_right_series = _order_series_based_on_node(right_node, right_series)
    return ordered_left_series, ordered_right_series


def _order_series_based_on_node(
        original_node: BachSqlModel,
        unordered_series: Mapping[str, Series]
) -> Dict[str, Series]:
    """
    Sort the given mapping of Series name to Series in the same way as the columns are sorted in the
    original_node. Series that are not in original_node are sorted at the end.

    :return: A sorted dictionary mapping Series.name to Series
    """
    ordered_series = {}
    for series_name in original_node.series_names:
        if series_name in unordered_series:
            ordered_series[series_name] = unordered_series[series_name]
    for series_name, series in unordered_series.items():
        if series_name not in ordered_series:
            ordered_series[series_name] = series
    return ordered_series


def _get_merge_sql_model(
    dialect: Dialect,
    left: DataFrame,
    right: DataFrameOrSeries,
    how: How,
    merge_on: MergeOn,
    new_column_list: List[ResultSeries],
    variables: Dict['DtypeNamePair', Hashable]
) -> BachSqlModel:
    """
    Give the SqlModel to join left and right and select the new_column_list. This model also uses the
    join-type of how, matching rows on real_left_on and real_right_on.
    """
    if merge_on.is_empty:
        on_clause = Expression.construct('')
    else:
        on_clause = _get_merge_on_clause(dialect, left, right, merge_on)

    columns_expr = join_expressions(
        [Expression.construct_expr_as_sql_name(dialect=dialect, expr=rc.expression, name=rc.name)
         for rc in new_column_list]
    )
    join_type_expr = Expression.construct('full outer' if how == How.outer else how.value)

    column_expressions = {rc.name: rc.expression for rc in new_column_list}
    return MergeSqlModel.get_instance(
        column_expressions=column_expressions,
        dialect=dialect,
        columns_expr=columns_expr,
        join_type_expr=join_type_expr,
        on_clause=on_clause,
        merge_on=merge_on,
        left_node=left.base_node,
        right_node=right.base_node,
        variables=variables,
    )


def _get_merge_on_clause(
    dialect: Dialect,
    left: DataFrame,
    right: DataFrameOrSeries,
    merge_on: MergeOn,
) -> Expression:
    """
    Generates the expression used as criteria to match rows when merging.
    """
    left_x_right_on_merge_expressions = list(itertools.chain.from_iterable(
        [
            _get_expression(df_series=left, label=l_label).resolve_column_references(
                dialect, _LEFT_NODE_ALIAS,
            ),
            _get_expression(df_series=right, label=r_label).resolve_column_references(
                dialect, _RIGHT_NODE_ALIAS,
            ),
        ]
        for l_label, r_label in zip(merge_on.left, merge_on.right)
    ))

    conditional_merge_expressions = _resolve_merge_expression_references(dialect, left, right, merge_on)

    all_conditions = (
        ['({} = {})'] * (len(left_x_right_on_merge_expressions) // 2)
        + ['({})'] * len(conditional_merge_expressions)
    )
    fmt_str = 'on ' + ' and '.join(all_conditions)

    return Expression.construct(
        fmt_str, *left_x_right_on_merge_expressions, *conditional_merge_expressions,
    )


def _resolve_merge_expression_references(
    dialect: Dialect,
    left: DataFrame,
    right: DataFrameOrSeries,
    merge_on: MergeOn
) -> List[Expression]:
    expressions = []
    df_series: DataFrameOrSeries
    for cond in merge_on.conditional:
        new_tokens: List[Union[ExpressionToken, 'Expression']] = []
        for token in cond.expression.get_all_tokens():
            if not isinstance(token, ColumnReferenceToken):
                new_tokens.append(token)
                continue

            token_original_name = _get_series_name_from_column_name(
                model=cond.base_node,
                dialect=dialect,
                column_name=token.column_name
            )
            prev_expression = cond.base_node.column_expressions[token_original_name]
            resolved_nested_tokens: List[Union[Expression, ExpressionToken]] = []

            for nested_token in prev_expression.get_all_tokens():
                if not isinstance(nested_token, TableColumnReferenceToken):
                    resolved_nested_tokens.append(nested_token)
                    continue

                ref_name = 'left_node' if nested_token.table_name == _LEFT_NODE_ALIAS else 'right_node'
                ref_node = cond.base_node.references[ref_name]

                if ref_node == left.base_node:
                    df_series = left
                    table_alias = _LEFT_NODE_ALIAS
                else:
                    df_series = right
                    table_alias = _RIGHT_NODE_ALIAS

                # nested_token.column_name might not be the same to token.column_name
                # since series can be renamed
                column_name = (
                    nested_token.column_name
                    if nested_token.column_name == get_sql_column_name(dialect, cond.name)
                    else token.column_name
                )
                expr_label = _get_series_name_from_column_name(
                    model=cond.base_node,
                    dialect=dialect,
                    column_name=column_name
                )
                resolved_expr = _get_expression(df_series=df_series, label=expr_label)
                resolved_expr = resolved_expr.resolve_column_references(dialect, table_alias)
                resolved_nested_tokens.append(resolved_expr)

            new_tokens.extend(resolved_nested_tokens)

        expressions.append(Expression(new_tokens))

    return expressions


def _get_series_name_from_column_name(model: BachSqlModel, dialect: Dialect, column_name: str) -> str:
    """
    Given a column_name and the BachSqlModel that the column appears in, determine what the series-name
    of the column name was. This is done by converting all series names in the BachSqlModel meta-data to
    column names, and seeing which matches the given column name.
    """
    series_names = list(model.column_expressions.keys())
    for series_name in series_names:
        if column_name == get_sql_column_name(dialect=dialect, name=series_name):
            return series_name
    raise ValueError(f'No Series name found matching column name "{column_name}" in series names '
                     f'{series_names}.')


def _get_expression(df_series: DataFrameOrSeries, label: str) -> Expression:
    """ Helper of merge: give the expression for the column with the given label in df_series """
    if df_series.index and label in df_series.index:
        return df_series.index[label].expression
    if isinstance(df_series, DataFrame):
        return df_series.data[label].expression
    if isinstance(df_series, Series):
        return df_series.expression
    raise TypeError(f'df_series should be DataFrameOrSeries. type: {type(df_series)}')


class MergeSqlModel(BachSqlModel):
    def __init__(self, merge_on: MergeOn, *args, **kwargs):
        self._merge_on = merge_on
        super().__init__(*args,  **kwargs)

    @property
    def merge_on(self) -> MergeOn:
        return self._merge_on

    @classmethod
    def get_instance(
        cls,
        *,
        dialect: Dialect,
        column_expressions: Dict[str, Expression],
        columns_expr: Expression,
        join_type_expr: Expression,
        on_clause: Expression,
        merge_on: MergeOn,
        left_node: BachSqlModel,
        right_node: BachSqlModel,
        variables: Dict['DtypeNamePair', Hashable],
    ) -> 'MergeSqlModel':
        """
        :param column_expressions: mapping between column_names and their expressions (in order)
        :param columns_expr: A single expression that expresses projecting all needed columns from either
            left or right
        :param join_type_expr: expression expressing the join type, e.g. an expression that represents
            the string 'inner', 'cross', or similar
        :param on_clause: single expression that expresses the on clause
        :param left_node, sql-model of the materialized left side of the join
        :param right_node, sql-model of the materialized right side of the join
        :param variables: Dictionary of all variable values
        """
        columns_str = columns_expr.to_sql(dialect)
        join_type_str = join_type_expr.to_sql(dialect)
        on_str = on_clause.to_sql(dialect)

        sql = f'''
            select {columns_str}
            from {{{{left_node}}}} as l {join_type_str}
            join {{{{right_node}}}} as r {on_str}
            '''
        name = 'merge_sql'

        # Add all references found in the Expressions to self.references
        all_expressions = [columns_expr, join_type_expr, on_clause]
        references = construct_references(
            base_references={'left_node': left_node, 'right_node': right_node},
            expressions=all_expressions
        )

        return MergeSqlModel(
            model_spec=CustomSqlModelBuilder(sql=sql, name=name),
            placeholders=cls._get_placeholders(dialect, variables, all_expressions),
            references=references,
            materialization=Materialization.CTE,
            materialization_name=None,
            column_expressions=column_expressions,
            merge_on=merge_on,
        )

    def copy_override(
        self: 'MergeSqlModel',
        *,
        model_spec: SqlModelSpec = None,
        placeholders: Mapping[str, Hashable] = None,
        references: Mapping[str, 'SqlModel'] = None,
        materialization: Materialization = None,
        materialization_name: Union[Optional[str], NotSet] = not_set,
        column_expressions: Dict[str, Expression] = None,
        merge_on: MergeOn = None
    ) -> 'MergeSqlModel':
        """
        Similar to super class's implementation, but adds optional 'merge_on' parameter
        """
        materialization_name_value = (
            self.materialization_name if materialization_name is not_set else materialization_name
        )
        return self.__class__(
            model_spec=self.model_spec if model_spec is None else model_spec,
            placeholders=self.placeholders if placeholders is None else placeholders,
            references=self.references if references is None else references,
            materialization=self.materialization if materialization is None else materialization,
            materialization_name=materialization_name_value,
            column_expressions=self.column_expressions if column_expressions is None else column_expressions,
            merge_on=self.merge_on if merge_on is None else merge_on
        )



"""
Copyright 2021 Objectiv B.V.
"""
import json
import operator
from functools import reduce
from typing import Dict, Union, TYPE_CHECKING, Tuple, cast, Optional, List, Any, TypeVar, Generic

from sqlalchemy.engine import Dialect

from bach import SortColumn
from bach.series import Series
from bach.expression import Expression, join_expressions
from bach.series.series import WrappedPartition, ToPandasInfo, value_to_series
from bach.sql_model import BachSqlModel
from bach.types import DtypeOrAlias, StructuredDtype, AllSupportedLiteralTypes
from sql_models.constants import DBDialect
from sql_models.model import Materialization
from sql_models.util import quote_string, is_postgres, DatabaseNotSupportedException, is_bigquery, is_athena

if TYPE_CHECKING:
    from bach.series import SeriesBoolean, SeriesString, SeriesInt64
    from bach.partitioning import GroupBy


class SeriesJson(Series):
    """
    A Series that represents the JSON type and its specific operations.


    **Database support and types**

    * On Postgres this utilizes the native 'jsonb' database type.
    * On Athena this utilizes the native 'json' database type.
    * On BigQuery this utilizes the generic 'STRING' database type.

    .. note::
        On Postgres, SeriesJson does not use the 'json' database type, but the 'jsonb' type, As the 'json'
        type has limited functionality.

        This class is the standard and recommended type to use for handling json like data. Having said that,
        there is a special SeriesJsonPostgres type that uses the 'json' data type on Postgres, but internally
        that casts all data to 'jsonb' too.

    **Getting data**

    It is possible to get a selection of data from the json in the json type column. For selecting data from
    json, arrays and objects are supported. The data can be selected using `.json[]` on the json column

    Selecting data from an array is based on position. It works similar to slicing through python lists.

    .. note::

        Slicing is only possible if *all* values in the column are lists or None.

    Selecting from objects is possible by key.

    Examples:

     .. testsetup:: json
        :skipif: engine is None

        data = ['["a","b","c"]', '["d","e","f","g"]', '[{"h":"i","j":"k"},{"l":["m","n","o"]},{"p":"q"}]']
        pdf = pd.DataFrame(data=data, columns=['jsonb_column'])
        df = DataFrame.from_pandas(engine, pdf)
        df['jsonb_column'] = df.jsonb_column.astype('json')

    .. doctest:: json
        :skipif: engine is None

        >>> pdf
                                                jsonb_column
        0                                      ["a","b","c"]
        1                                  ["d","e","f","g"]
        2  [{"h":"i","j":"k"},{"l":["m","n","o"]},{"p":"q"}]

    .. doctest:: json
        :skipif: engine is None

        >>> df = DataFrame.from_pandas(engine, pdf)
        >>> df['jsonb_column'] = df.jsonb_column.astype('json')
        >>> # load some json strings and convert them to jsonb type
        >>> # slice and show with .head()
        >>> df.jsonb_column.json[:2].head()
        _index_0
        0                                            [a, b]
        1                                            [d, e]
        2    [{'h': 'i', 'j': 'k'}, {'l': ['m', 'n', 'o']}]
        Name: jsonb_column, dtype: object

    .. doctest:: json
        :skipif: engine is None

        >>> df.jsonb_column.json[1].head()
        _index_0
        0                         b
        1                         e
        2    {'l': ['m', 'n', 'o']}
        Name: jsonb_column, dtype: object

    .. doctest:: json
        :skipif: engine is None

        >>> # selecting from objects is done by entering a key:
        >>> df.jsonb_column.json[1].json['l'].head()
        _index_0
        0         None
        1         None
        2    [m, n, o]
        Name: jsonb_column, dtype: object

    A last case is selecting based on the objects *in* an array.
    With this method, a dict is passed in the `.json[]` selector. The value of the first match with the dict
    to the objects in a json array is returned for the `.json[]` selector. A match is when all key/value pairs
    of the dict are found in an object. This can be used for selecting a subset of a json array with objects.

    .. doctest:: json
        :skipif: engine is None

        >>> # selecting from arrays by searching objects in the array.
        >>> df.jsonb_column.json[:{"j":"k"}].head()
        _index_0
        0                        []
        1                        []
        2    [{'h': 'i', 'j': 'k'}]
        Name: jsonb_column, dtype: object

    .. doctest:: json
        :skipif: engine is None

        >>> # or:
        >>> df.jsonb_column.json[{"l":["m","n","o"]}:].head()
        _index_0
        0                                      []
        1                                      []
        2    [{'l': ['m', 'n', 'o']}, {'p': 'q'}]
        Name: jsonb_column, dtype: object
    """

    # TODO: support instance_dtype to return better types
    dtype = 'json'
    dtype_aliases: Tuple[DtypeOrAlias, ...] = tuple()
    supported_db_dtype = {
        DBDialect.POSTGRES: 'jsonb',
        DBDialect.ATHENA: 'json',
        # on BigQuery we use STRING for JSONs. Of course SeriesString is the default class for that, so
        # here we set None
        DBDialect.BIGQUERY: None
    }
    supported_value_types = (dict, list, str, float, int, bool)

    supported_source_dtypes = ('json', 'json_postgres', 'string')

    @property
    def json(self) -> 'JsonAccessor':
        """
        Get access to json operations via the class that's returned through this accessor.
        Use as `my_series.json.get_value()` or `my_series.json[:2]`

        .. autoclass:: bach.series.series_json.JsonAccessor
            :members:

        """
        return JsonAccessor(series_object=self)

    @property
    def elements(self):
        return self.json

    @classmethod
    def get_db_dtype(cls, dialect: Dialect) -> Optional[str]:
        if is_bigquery(dialect):
            from bach.series import SeriesString
            return SeriesString.get_db_dtype(dialect)
        return super().get_db_dtype(dialect)

    @classmethod
    def supported_literal_to_expression(cls, dialect: Dialect, literal: Expression) -> Expression:
        if is_postgres(dialect):
            return super().supported_literal_to_expression(dialect=dialect, literal=literal)
        if is_athena(dialect):
            return Expression.construct('json_parse({})', literal)
        if is_bigquery(dialect):
            from bach import SeriesString
            return SeriesString.supported_literal_to_expression(dialect=dialect, literal=literal)
        raise DatabaseNotSupportedException(dialect)

    @classmethod
    def supported_value_to_literal(
            cls,
            dialect: Dialect,
            value: Union[dict, list],
            dtype: StructuredDtype
    ) -> Expression:
        cls.assert_engine_dialect_supported(dialect)
        json_value = json.dumps(value)
        return Expression.string_value(json_value)

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        if is_postgres(dialect):
            if source_dtype in ('json', 'json_postgres'):
                # SeriesJsonPostgres is a special case: SeriesJsonPostgres.expression already contains a cast
                # to the database type 'jsonb', so we actually don't need to do any conversion
                return expression
            if source_dtype == 'string':
                return Expression.construct(f'cast({{}} as {cls.get_db_dtype(dialect)})', expression)
            raise ValueError(f'cannot convert {source_dtype} to json')
        if is_athena(dialect):
            if source_dtype == 'json':
                return expression
            if source_dtype == 'string':
                return Expression.construct('json_parse({})', expression)
            raise ValueError(f'cannot convert {source_dtype} to json')
        if is_bigquery(dialect):
            if source_dtype in ('json', 'string'):
                return expression
            elif source_dtype in ('dict', 'list'):
                # BQ can convert most types into json format, and since we use string as the container
                # in our BQ implementation, this is basically a cast.
                # https://cloud.google.com/bigquery/docs/reference/standard-sql/json_functions#to_json_string
                return Expression.construct(f'to_json_string({{}})', expression)
            else:
                raise ValueError(f'cannot convert {source_dtype} to json')
        raise DatabaseNotSupportedException(dialect)

    def to_pandas_info(self) -> Optional['ToPandasInfo']:
        if is_bigquery(self.engine):
            # All data is stored as string, so if we actually want the objects, we need to load the string
            # as json.
            return ToPandasInfo('object', self._json_loads)
        return None

    @staticmethod
    def _json_loads(data: Optional[str]) -> Optional[Any]:
        """ Helper of to_pandas_info """
        if data is None:
            return None
        try:
            return json.loads(data)
        except ValueError as exc:
            raise ValueError(f'invalid json content in result: {data}') from exc

    def _comparator_operation(
        self,
        other: Union['Series', AllSupportedLiteralTypes],
        comparator: str,
        other_dtypes=tuple(),
        strict_other_dtypes=tuple()
    ) -> 'SeriesBoolean':
        if is_postgres(self.engine):
            other_dtypes = ('json_postgres', 'json', 'string')
            strict_other_dtypes = tuple(['json_postgres'])
        elif is_athena(self.engine):
            other_dtypes = ('json', 'string')
            strict_other_dtypes = tuple(['json'])
        elif is_bigquery(self.engine):
            other_dtypes = ('json', 'string')
        else:
            raise DatabaseNotSupportedException(self.engine)

        result = self._binary_operation(
            other, operation=f"comparator '{comparator}'",
            fmt_str=f'{{}} {comparator} {{}}',
            other_dtypes=other_dtypes, dtype='bool', strict_other_dtypes=strict_other_dtypes
        )
        return cast('SeriesBoolean', result)  # we told _binary_operation to return dtype='bool'

    def __le__(self, other: Union['Series', AllSupportedLiteralTypes]) -> 'SeriesBoolean':
        raise NotImplementedError()

    def __ge__(self, other: Union['Series', AllSupportedLiteralTypes]) -> 'SeriesBoolean':
        raise NotImplementedError()

    def min(self, partition: WrappedPartition = None, skipna: bool = True):
        """ INTERNAL: Only here to not trigger errors from describe """
        raise NotImplementedError()

    def max(self, partition: WrappedPartition = None, skipna: bool = True):
        """ INTERNAL: Only here to not trigger errors from describe """
        raise NotImplementedError()


class SeriesJsonPostgres(SeriesJson):
    """
    A special Series that represents the 'json' database type on Postgres.

    .. note::
        Generally, it is advised to use :class:`JsonSeries` instead. Given a `SeriesJsonPostgres` object,
        calling `json_series_pg.astype('json')` will return a `JsonSeries` class representing the same data.

    When `json` data is encountered in a sql table, this dtype is used. On Postgres for all operations the
    data is first cast to jsonb type. Therefore, it is recommended to cast to :class:`JsonSeries` directly.

    The public interface of this class is the same as the :py:class:`SeriesJson` class. See the docstring
    of that class for more information.
    """

    dtype = 'json_postgres'
    dtype_aliases: Tuple[DtypeOrAlias, ...] = tuple()
    supported_db_dtype = {
        DBDialect.POSTGRES: 'json',
    }

    def __init__(self,
                 engine,
                 base_node: BachSqlModel,
                 index: Dict[str, 'Series'],
                 name: str,
                 expression: Expression,
                 group_by: 'GroupBy',
                 order_by: Optional[List[SortColumn]],
                 instance_dtype: StructuredDtype,
                 **kwargs):

        super().__init__(engine=engine,
                         base_node=base_node,
                         index=index,
                         name=name,
                         expression=Expression.construct(f'cast({{}} as jsonb)', expression),
                         group_by=group_by,
                         order_by=order_by,
                         instance_dtype=instance_dtype,
                         **kwargs)

    @property
    def json(self) -> 'JsonAccessor':
        """
        Get access to json operations via the class that's returned through this accessor.
        Use as `my_series.json.get_value()` or `my_series.json[:2]`

        .. autoclass:: bach.series.series_json.JsonAccessor
            :members:
        """
        json_series = self.astype('json')
        assert isinstance(json_series, SeriesJson)  # help mypy
        return json_series.json

    def materialize(
            self,
            node_name='manual_materialize',
            limit: Any = None,
            distinct: bool = False,
            materialization: Union[Materialization, str] = Materialization.CTE
    ):
        """
        Instance of this Series cannot be materialized, as the base expression is not just the column name
        but a cast to 'jsonb'.

        Use class:`SeriesJson` instead, which can be materialized.
        """
        raise Exception(f'{self.__class__.__name__} cannot be materialized.')


TSeriesJson = TypeVar('TSeriesJson', bound='SeriesJson')


class JsonAccessor(Generic[TSeriesJson]):
    """
    Class with accessor methods to JSON type data.
    """

    # We have different implementations for different databases. This class is a facade, that will call the
    # actual implementing class based on the used database engine.
    # Having this facade (instead of having subclasses per engine) makes it possible to have classes inherit
    # from this class.

    def __init__(self, series_object: 'TSeriesJson'):
        self._series_object = series_object

        # _implementation implements the actual functionality.
        engine = self._series_object.engine
        self._implementation: Union[
            JsonPostgresAccessorImpl, JsonAthenaAccessorImpl, JsonBigQueryAccessorImpl
        ]
        if is_postgres(engine):
            self._implementation = JsonPostgresAccessorImpl(series_object)
        elif is_athena(engine):
            self._implementation = JsonAthenaAccessorImpl(series_object)
        elif is_bigquery(engine):
            self._implementation = JsonBigQueryAccessorImpl(series_object)
        else:
            raise DatabaseNotSupportedException(engine)

    def __getitem__(self, key: Union[str, int, slice]) -> 'TSeriesJson':
        """
        Get item(s) from the JSON data in pythonic ways.
        The way item(s) are looked up depends on the type of key.

        If key is an integer, then this returns an item from the json array.
        The key is treated as a 0-based index. If negative this will count from the end of the array (one
            based). If the index does not exists this will render None/NULL.
        This assumes the top-level item in the json is an array

        If key is a slice, then this returns a slice from the json array.
        This assumes the top-level item in the json is an array

        If key is a string, then this returns an item from the json object. The item belonging to the given
            key is returned, or None/NULL if not found.
        This assumes the top-level item in the json is an object
        """
        # TODO: leverage instance_dtype information here, if we have that
        if isinstance(key, int):
            return self._implementation.get_array_item(key)

        elif isinstance(key, str):
            return self._implementation.get_dict_item(key)

        elif isinstance(key, slice):
            return self._implementation.get_array_slice(key)

        raise TypeError('Key should either be a string, integer, or slice.')

    def get_value(self, key: str, as_str: bool = False) -> Union['SeriesString', 'TSeriesJson']:
        """
        Get item from toplevel object by key.

        :param key: the key to return the values for.
        :param as_str: if True, it returns a string Series, json otherwise. Particular useful if the returned
            value is in fact a string. In json the string will be represented as a quoted string, if this
            field is set to true, the returned SeriesString will not have additional quotes.
        :returns: series with the selected object value.
        """
        return self._implementation.get_value(key=key, as_str=as_str)

    def get_array_length(self) -> 'SeriesInt64':
        """
        Get the length of the toplevel array.

        This assumes the top-level item in the json is an array. Will result in an exception (later on) if
        that's not the case!
        """
        # Implementing a more generic __len__() function is not trivial as a json object can be (among
        # others) an array, a dict, or a string, all of which should be supported by a generic __len__().
        # So for now we have a dedicated len function for arrays.
        return self._implementation.get_array_length()

    def array_contains(self, item: Union[int, float, bool, str, None]) -> 'SeriesBoolean':
        """
        Find if item is contained in the array.

        :param item: item to be verified if it is contained by the array. Only supports scalar values.
        :returns: boolean series indicating if the array contains the element.

        This assumes the top-level item in the json is an array.
        """

        return self._implementation.array_contains(item)

    def flatten_array(self) -> Tuple['TSeriesJson', 'SeriesInt64']:
        """
        Converts elements in an array into a set of rows.

        Since the operation might destroy the order of the elements,
        a series containing the offset is also returned.

        :returns: Tuple with SeriesJson (element from the array) and SeriesInt64 (0-based offset of the
            element)

        .. note::
            Both returned series objects will share same base node, but this node is not
            the same as the unflatten Series.

        This assumes the top-level item in the json is an array.
        """
        return self._implementation.flatten_array()


class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
    """
    BigQuery specific implementation of JsonAccessor functions.
    """

    def __init__(self, series_object: 'TSeriesJson'):
        self._series_object = series_object

    def get_array_slice(self, key: slice) -> 'TSeriesJson':
        """ See implementation in parent class :class:`JsonAccessor` """
        if key.step:
            raise NotImplementedError('slice steps not supported')

        start_expression = self._get_slice_partial_expr(value=key.start, is_start=True)
        stop_expression = self._get_slice_partial_expr(value=key.stop, is_start=False)

        values_expression = Expression.construct(
            "select val "
            "from unnest(JSON_QUERY_ARRAY({}, '$')) val with offset as pos "
            "where pos >= {} and pos < {} "
            "order by pos",
            self._series_object, start_expression, stop_expression
        )
        json_str_expression = Expression.construct(
            "'[' || ARRAY_TO_STRING(ARRAY({}), ', ') || ']'",
            values_expression
        )
        return self._series_object\
            .copy_override(expression=json_str_expression)

    def _get_slice_partial_expr(
        self,
        value: Optional[Union[Dict[str, str], int]],
        is_start: bool,
    ) -> Expression:
        """
        Helper of get_array_slice(). return expression for either the lower bound or upper bound of a slice.

        Assumes that self._series_object is an array!

        :param value: slice.start or slice.stop
        :param is_start: whether value is the slice.start (True) or slice.stop (False) value
        :return: Expression that will evaluate to an integer that can be used to compare against positions
                    in the array.
        """
        if isinstance(value, dict):
            return self._find_in_json_list(value, is_start)

        if value is not None and not isinstance(value, int):
            raise TypeError(f'Slice value must be None or an integer, value: {value}')

        if value is None:
            if is_start:
                # return index of first item in the array
                return Expression.construct('0')
            else:
                # return index that is guaranteed to be beyond the last item in the array
                max_int = 2 ** 63 - 1
                return Expression.construct(f'{max_int}')
        elif value >= 0:
            return Expression.construct(f'{value}')
        else:
            array_len = self.get_array_length()
            return Expression.construct(f'({{}} {value})', array_len)

    def _find_in_json_list(self, filtering_slice: Dict[str, str], is_start: bool):
        """
        Return expression for the position of the element that contains the filtering values

        Assumes that self._series_object is an array and each element is a json!

        :param filtering_slice: dictionary containing the keys and values to use on filter.
        :param is_start: whether value is the slice.start (True) or slice.stop (False) value. If true,
            the first element containing the values is returned, otherwise the last.
        :return: Expression that gets the position of the element based on the slicing filters.
        """
        if not isinstance(filtering_slice, dict):
            raise TypeError(f'key should be a dict, actual type: {type(filtering_slice)}')

        # this way we can reuse code from the accessor without duplicating.
        # 'element' will be filled by iterating over the json array in the query below.
        element_json = self._series_object.copy_override(
            expression=Expression.construct('element'),
        )

        all_filters = [
            element_json.json.get_value(filter_key, as_str=True) == str(filter_value)
            for filter_key, filter_value in filtering_slice.items()
        ]
        slicing_mask = reduce(operator.and_, all_filters)
        fmt = (
            # end slicing is inclusive
            f"(select {'min' if is_start else '1 + max'}(case when {{}} then pos end) "
            f"from unnest(JSON_QUERY_ARRAY({{}}, '$')) element with offset as pos)"
        )
        return Expression.construct(fmt, slicing_mask, self._series_object)

    def get_array_item(self, key: int) -> 'TSeriesJson':
        """
        Returns an item from the json array.
        The key is treated as a 0-based index. If negative this will count from the end of the array (one
            based). If the index does not exist this will render None/NULL.
        This assumes the top-level item in the json is an array
        """
        if key < 0:
            # BigQuery doesn't (yet) natively support this, so we emulate this by reversing the array
            key = abs(key)
            expression = Expression.construct(
                f'ARRAY_REVERSE(JSON_QUERY_ARRAY({{}}))[SAFE_ORDINAL({key})]',
                self._series_object
            )
        else:
            expression = Expression.construct(f'''JSON_QUERY({{}}, '$[{key}]')''', self._series_object)
        return self._series_object.copy_override(expression=expression)

    def get_dict_item(self, key: str) -> 'TSeriesJson':
        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
        return cast('TSeriesJson', self.get_value(key=key, as_str=False))

    def get_value(self, key: str, as_str: bool = False) -> Union['SeriesString', 'TSeriesJson']:
        """ For documentation, see function :meth:`JsonAccessor.get_value()` """
        # Special characters are tricky. According to the latest jsonpath spec draft we should use the
        # index selector [1] (e.g. `$['key with special characters']`). But BigQuery only supports the dot
        # selector [2][3]. However BigQuery does allow us to quote the key when using the dot selector [4],
        # which is not in the jsonpath draft spec. So we quote the key, and just raise an exception if
        # anybody tries to use a key that contains a quote.
        # [1] https://www.ietf.org/archive/id/draft-ietf-jsonpath-base-05.html#name-index-selector
        # [2] https://www.ietf.org/archive/id/draft-ietf-jsonpath-base-05.html#name-dot-selector
        # [3] https://cloud.google.com/bigquery/docs/reference/standard-sql/json_functions#JSONPath_format
        # [4] https://cloud.google.com/bigquery/docs/reference/standard-sql/json_functions
        if '"' in key:
            raise ValueError(f'key values containing double quotes are not supported. key: {key}')

        json_expression = Expression.construct(
            '''JSON_QUERY({}, '$."{}"')''',
            self._series_object,
            Expression.raw(key)
        )

        if not as_str:
            return self._series_object.copy_override(expression=json_expression)
        from bach.series import SeriesString
        # JSON_QUERY will return a string containing valid json. i.e. if the queried value is a string, then
        # it will be quoted. Here we remove the quotes
        value_expression = Expression.construct(
            f'REGEXP_REPLACE({{}}, {{}}, {{}})',
            json_expression,
            Expression.raw('r\'(^"|"$)\''),
            Expression.string_value(''),
        )
        return self._series_object.copy_override(
            expression=value_expression
        ).copy_override_type(SeriesString)

    def get_array_length(self) -> 'SeriesInt64':
        """ For documentation, see function :meth:`JsonAccessor.get_array_length()` """
        # Implementing a more generic __len__() function is not trivial as a json object can be (among
        # others) an array, a dict, or a string, all of which should be supported by a generic __len__().
        # So for now we have a dedicated len function for arrays.
        from bach.series import SeriesInt64
        expression = Expression.construct('ARRAY_LENGTH(JSON_QUERY_ARRAY({}))', self._series_object)
        return self._series_object \
            .copy_override_type(SeriesInt64) \
            .copy_override(expression=expression)

    def array_contains(self, item: Union[int, float, bool, str, None]) -> 'SeriesBoolean':
        """ For documentation, see implementation in class :class:`JsonAccessor` """
        item_json = json.dumps(item)
        in_expr = Expression.construct(
            '(SELECT {} IN UNNEST(JSON_QUERY_ARRAY({})))',
            Expression.string_value(item_json),
            self._series_object
        )
        expression = Expression.construct('IF({} IS NULL, NULL, {})', self._series_object, in_expr)
        from bach.series import SeriesBoolean
        return self._series_object\
            .copy_override(expression=expression)\
            .copy_override_type(SeriesBoolean)

    def flatten_array(self) -> Tuple['TSeriesJson', 'SeriesInt64']:
        """ For documentation, see implementation in class :class:`JsonAccessor` """
        from bach.series.array_operations.flattening import BigQueryArrayFlattening
        return BigQueryArrayFlattening(self._series_object)()


class JsonPostgresAccessorImpl(Generic[TSeriesJson]):
    """
    Postgres specific implementation of JsonAccessor functions.
    """

    def __init__(self, series_object: 'TSeriesJson'):
        self._series_object = series_object

    def _find_in_json_list(self, key: Union[str, Dict[str, str]]):
        if isinstance(key, (dict, str)):
            key = json.dumps(key)
            quoted_key = quote_string(self._series_object.engine, key)
            expression_str = f"""(select min(case when ({quoted_key}::jsonb) <@ value
            then ordinality end) -1 from jsonb_array_elements({{}}) with ordinality)"""
            return expression_str
        else:
            raise TypeError(f'key should be int or slice, actual type: {type(key)}')

    def get_array_slice(self, key: slice) -> 'TSeriesJson':
        """ See implementation in parent class :class:`JsonAccessor` """
        expression_references = 0
        if key.step:
            raise NotImplementedError('slice steps not supported')
        if key.stop is not None:
            negative_stop = ''
            if isinstance(key.stop, int):
                if key.stop < 0:
                    negative_stop = f'jsonb_array_length({{}})'
                    expression_references += 1
                stop = f'{negative_stop} {key.stop} - 1'
            elif isinstance(key.stop, (dict, str)):
                stop = self._find_in_json_list(key.stop)
                expression_references += 1
            else:
                raise TypeError('cant')
        if key.start is not None:
            if isinstance(key.start, int):
                negative_start = ''
                if key.start < 0:
                    negative_start = f'jsonb_array_length({{}})'
                    expression_references += 1
                start = f'{negative_start} {key.start}'
            elif isinstance(key.start, (dict, str)):
                start = self._find_in_json_list(key.start)
                expression_references += 1
            else:
                raise TypeError('cant')
            if key.stop is not None:
                where = f'between {start} and {stop}'
            else:
                where = f'>= {start}'
        else:
            if key.stop is not None:
                where = f'<= {stop}'
            else:
                # no start and no stop: we want to select all elements.
                where = 'is not null'  # should be true for all ordinalities.
        combined_expression = f"""(select jsonb_agg(x.value)
        from jsonb_array_elements({{}}) with ordinality x
        where ordinality - 1 {where})"""
        expression_references += 1
        non_null_expression = f"coalesce({combined_expression}, '[]'::jsonb)"
        return self._series_object \
            .copy_override(
                expression=Expression.construct(
                    non_null_expression,
                    *([self._series_object] * expression_references)
                )
            )

    def get_array_item(self, key: int) -> 'TSeriesJson':
        """
        Returns an item from the json array.
        The key is treated as a 0-based index. If negative this will count from the end of the array (one
            based). If the index does not exist this will render None/NULL.
        This assumes the top-level item in the json is an array
        """
        return self._series_object \
            .copy_override(expression=Expression.construct(f'{{}}->{key}', self._series_object))

    def get_dict_item(self, key: str) -> 'TSeriesJson':
        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
        return cast('TSeriesJson', self.get_value(key=key, as_str=False))

    def get_value(self, key: str, as_str: bool = False) -> Union['SeriesString', 'TSeriesJson']:
        """ For documentation, see function :meth:`JsonAccessor.get_value()` """

        if '"' in key:
            raise ValueError(f'key values containing double quotes are not supported. key: {key}')

        return_as_string_operator = ''
        if as_str:
            return_as_string_operator = '>'
        expression = Expression.construct(f"{{}}->{return_as_string_operator}{{}}",
                                          self._series_object,
                                          Expression.string_value(key))
        result = self._series_object.copy_override(expression=expression)
        if as_str:
            from bach.series import SeriesString
            return result.astype('string').copy_override_type(SeriesString)
        return result

    def get_array_length(self) -> 'SeriesInt64':
        """ For documentation, see function :meth:`JsonAccessor.get_array_length()` """
        from bach.series import SeriesInt64
        expression = Expression.construct('jsonb_array_length({})', self._series_object)
        return self._series_object \
            .copy_override_type(SeriesInt64) \
            .copy_override(expression=expression)

    def array_contains(self, item: Union[int, float, bool, str, None]) -> 'SeriesBoolean':
        """ For documentation, see implementation in class :class:`JsonAccessor` """
        return self._series_object._comparator_operation([item], "@>")

    def flatten_array(self) -> Tuple['TSeriesJson', 'SeriesInt64']:
        """ For documentation, see implementation in class :class:`JsonAccessor` """
        from bach.series.array_operations.flattening import PostgresArrayFlattening
        return PostgresArrayFlattening(self._series_object)()


class JsonAthenaAccessorImpl(Generic[TSeriesJson]):
    """
    Athena specific implementation of JsonAccessor functions.
    """

    def __init__(self, series_object: 'TSeriesJson'):
        self._series_object = series_object

    def get_array_slice(self, key: slice) -> 'TSeriesJson':
        """ See implementation in parent class :class:`JsonAccessor` """
        if key.step:
            raise NotImplementedError('slice steps not supported')

        start_expression = self._get_slice_partial_expr(value=key.start, is_start=True)
        stop_expression = self._get_slice_partial_expr(value=key.stop, is_start=False)

        values_expression = Expression.construct(
            "try(slice(cast({} as array(json)), {}, {}))",
            self._series_object,
            start_expression,  # startpoint of slice
            Expression.construct('({} - {})', stop_expression, start_expression),  # length of slice
        )
        non_null_expr = Expression.construct('coalesce({}, ARRAY [])', values_expression)
        json_expression = Expression.construct('cast({} as json)', non_null_expr)
        return self._series_object\
            .copy_override(expression=json_expression)

    def _get_slice_partial_expr(
        self,
        value: Optional[Union[Dict[str, str], int]],
        is_start: bool,
    ) -> Expression:
        """
        Helper of get_array_slice(). return expression for either the lower bound or upper bound of a slice.

        Assumes that self._series_object is an array!

        :param value: slice.start or slice.stop
        :param is_start: whether value is the slice.start (True) or slice.stop (False) value
        :return: Expression that will evaluate to an integer that can be used as an one-based array index.
        """
        if isinstance(value, dict):
            return self._find_in_json_list(value, is_start)

        if value is not None and not isinstance(value, int):
            raise TypeError(f'Slice value must be None or an integer, value: {value}')

        if value is None:
            if is_start:
                # return index of first item in the array
                return Expression.construct('1')
            else:
                # return index that is guaranteed to be beyond the last item in the array, but no bigger
                # than a 32-bit integer. Experiments have shown that slice() will give empty results, if
                # the length is larger than a 32 bit integer.
                max_int = 2 ** 31 - 1
                return Expression.construct(f'{max_int}')
        elif value >= 0:
            return Expression.construct(f'{value + 1}')
        else:
            array_len = self.get_array_length()
            return Expression.construct(f'({{}} {value} + 1)', array_len)

    def _find_in_json_list(self, filtering_slice: Dict[str, str], is_start: bool):
        """
        Return expression for the position of the element that contains the filtering values. The element
        can contain more values than the filtering values, but it must contain all of the values in
        filtering_slice.

        Assumes that self._series_object is an array, each element is a json object!

        :param filtering_slice: dictionary containing the keys and values to use on filter.
            Must be a dictionary with strings as keys and values!
        :param is_start: whether value is the slice.start (True) or slice.stop (False) value. If true,
            the first element containing the values is returned, otherwise the last.
        :return: Expression that gets the position (one-based) of the element based on the slicing filters.
        """
        # Athena doesn't support correlated subqueries, so we cannot use a select here to find the position
        # of the element.
        # Strategy:
        # 1. convert to array of maps
        # 2. filter array of maps on the values we look for
        # 3.1 for is_start:
        #   find first position of first element from the filtered array in the original json array
        # 3.2 for not is_start:
        #   find last position of last element from the filtered array in the original json array

        if not isinstance(filtering_slice, dict):
            raise TypeError(f'Key should be a dict, actual type: {type(filtering_slice)}')
        if not all(isinstance(key, str) for key in filtering_slice.keys()):
            raise TypeError(f'Keys in the key dict, should all be strings')
        if not all(isinstance(value, str) for value in filtering_slice.values()):
            raise TypeError(f'Values in the key dict, should all be strings')

        array_map_expr = Expression.construct(
            'try(cast({} as array(map(varchar, json))))',
            self._series_object
        )
        key_to_extract = [
            Expression.construct(
                '({}, __x[{}])',
                Expression.string_value(key),
                Expression.string_value(key),
            )
            for key, value in filtering_slice.items()
        ]
        key_values_to_match = [
            Expression.construct(
                '({}, {})',
                Expression.string_value(key),
                Expression.string_value(value),
            )
            for key, value in filtering_slice.items()
        ]

        # extracts only the keys to match
        transform_array = Expression.construct(
            'transform({}, __x -> cast(map_from_entries(array[{}]) as json))',
            array_map_expr,
            join_expressions(key_to_extract),
        )
        element_to_match = Expression.construct(
            'cast(map_from_entries(array[{}]) as json)',
            join_expressions(key_values_to_match)
        )
        if is_start:
            # Find the position in the original array, of the first object that matches the filter criteria
            first_element_pos_expr = Expression.construct(
                'array_position({}, {})',
                transform_array,
                element_to_match,
            )
            return first_element_pos_expr

        # elif not is_start: Find the position in the reversed transformed array, of the last element that
        # matches the filter criteria. This is the last element that matches the criteria in the original
        # array.
        # Additionally, we need to add 1, because for historical reasons the end of a filtering slice
        # includes the last matched element (unlike a regular slice where the end element is not
        # included).
        return Expression.construct(
            'array_position(reverse({}), {}) + 1',
            transform_array,
            element_to_match,
        )

    def get_array_item(self, key: int) -> 'TSeriesJson':
        """
        Returns an item from the json array.
        The key is treated as a 0-based index. If negative this will count from the end of the array (one
            based). If the index does not exist this will render None/NULL.
        This assumes the top-level item in the json is an array
        """
        # Athena also has a function `json_array_get()`. That function should do what we want, but it's
        # broken and of no use to us as it returns invalid json in some cases [1]. Instead, we cast the
        # json to a sql-array of jsons, and get the requested item from that.
        # [1] https://prestodb.io/docs/0.217/functions/json.html

        # sql-arrays use zero-based indexing for positive values.
        offset = key if key < 0 else key + 1
        expression = Expression.construct(
            f'element_at(try(cast({{}} as array(json))), {offset})',
            self._series_object
        )
        return self._series_object.copy_override(expression=expression)

    def get_dict_item(self, key: str) -> 'TSeriesJson':
        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
        return cast('TSeriesJson', self.get_value(key=key, as_str=False))

    def get_value(self, key: str, as_str: bool = False) -> Union['SeriesString', 'TSeriesJson']:
        """ For documentation, see function :meth:`JsonAccessor.get_value()` """
        if '"' in key:
            raise ValueError(f'key values containing double quotes are not supported. key: {key}')

        json_expression = Expression.construct(
            '''json_extract({}, '$["{}"]')''',
            self._series_object,
            Expression.raw(key)
        )

        if not as_str:
            return self._series_object.copy_override(expression=json_expression)

        # TODO: use SeriesString.str.replace() when it supports regexp
        str_expression = Expression.construct('json_format({})', json_expression)
        value_expression = Expression.construct(
            f'regexp_replace({{}}, {{}}, {{}})',
            str_expression,
            Expression.raw('\'(^"|"$)\''),
            Expression.string_value(''),
        )
        from bach.series import SeriesString
        return self._series_object\
            .copy_override(expression=value_expression)\
            .copy_override_type(SeriesString)

    def get_array_length(self) -> 'SeriesInt64':
        """ For documentation, see function :meth:`JsonAccessor.get_array_length()` """
        from bach.series import SeriesInt64
        expression = Expression.construct('json_array_length({})', self._series_object)
        return self._series_object \
            .copy_override_type(SeriesInt64) \
            .copy_override(expression=expression)

    def array_contains(self, item: Union[int, float, bool, str, None]) -> 'SeriesBoolean':
        """ For documentation, see implementation in class :class:`JsonAccessor` """
        from bach.series import SeriesBoolean
        if item is None:
            # json_array_contains does not work when the item we are looking for is `null`, it will always
            # return a sql NULL in that case. So for that special case we fall back to the regular array
            # contains.
            expression = Expression.construct(
                "contains(cast({} as array(json)), json_parse('null'))",
                self._series_object
            )
        else:
            # Regular case.
            # The second argument to json_array_contains has to be an Athena native bigint, double, bool, or
            # varchar, not a json.
            item_expr = value_to_series(base=self._series_object, value=item).expression
            expression = Expression.construct('json_array_contains({}, {})', self._series_object, item_expr)
        return self._series_object \
            .copy_override_type(SeriesBoolean) \
            .copy_override(expression=expression)

    def flatten_array(self) -> Tuple['TSeriesJson', 'SeriesInt64']:
        """ For documentation, see implementation in class :class:`JsonAccessor` """
        from bach.series.array_operations.flattening import AthenaArrayFlattening
        return AthenaArrayFlattening(self._series_object)()



from typing import Union, List

import pandas

from bach import DataFrame, SeriesFloat64
from bach.partitioning import WindowFrameBoundary, Window, GroupBy
from bach.series import SeriesAbstractNumeric, SeriesTimedelta
from bach.expression import Expression, AggregateFunctionExpression
from bach.series.series import WrappedPartition
from sql_models.util import is_bigquery, is_postgres, is_athena

_QUANTILES_SERIES_NAME = 'quantile'


def _get_valid_series_for_calculation(df: DataFrame) -> List[str]:
    """
    Returns data series names that support quantile operation.
    Currently supporting only numeric and timedelta series dtypes.
    """
    return [
        num_series.name
        for num_series in df.data.values()
        if isinstance(num_series, (SeriesAbstractNumeric, SeriesTimedelta))
    ]


def _add_quantiles_as_series(df: DataFrame, q: Union[float, List[float]]) -> DataFrame:
    """
    Adds a new series to the provided DataFrame containing all quantile values to be calculated.
    """
    quantiles_df = df.copy()

    if isinstance(q, float):
        quantiles_df[_QUANTILES_SERIES_NAME] = q
    else:
        quantiles_to_calculate = DataFrame.from_pandas(
            engine=df.engine,
            df=pandas.DataFrame({_QUANTILES_SERIES_NAME: q})
        )
        quantiles_to_calculate = quantiles_to_calculate.reset_index(drop=True)

        quantiles_df = quantiles_df.merge(quantiles_to_calculate, how='cross')

    return quantiles_df


def _calculate_quantiles_with_percentile_cont(
    df: DataFrame, q: Union[float, List[float]], group_by: List[str],
) -> DataFrame:
    """
    Helper function in charge of calculating quantiles using SQL's PERCENTILE_CONT
    function (if engine supports it).

    Supported only for Postgres and BigQuery.
    """
    if not (is_postgres(df.engine) or is_bigquery(df.engine)):
        raise Exception('Can only calculate quantiles using PERCENTILE_CONT for Postgres and Bigquery only.')

    quantiles_df = df.copy()
    window = None
    if is_bigquery(df.engine):
        window = quantiles_df.groupby(by=group_by).window(
            start_boundary=None, end_boundary=None,
        ).group_by
    else:
        quantiles_df = df.groupby(group_by)

    quantiles = [q] if isinstance(q, float) else q
    series_to_calculate = _get_valid_series_for_calculation(df)

    q_results = {}
    for qt in quantiles:
        for series_name in series_to_calculate:
            q_series_name = f'__{series_name}_{qt}_quantile'
            series_to_agg = quantiles_df[series_name]

            agg_expr = AggregateFunctionExpression.construct(
                f'percentile_cont({qt}) within group (order by {{}})',
                series_to_agg,
            )
            if is_bigquery(df.engine):
                # mypy: help
                if window is None or not isinstance(window, Window):
                    raise Exception('Invalid window object.')

                if isinstance(series_to_agg, SeriesTimedelta):
                    series_to_agg = series_to_agg.dt.total_seconds

                agg_expr = window.get_window_expression(
                    AggregateFunctionExpression.construct(
                        f'percentile_cont({{}}, {qt})', series_to_agg,
                    )
                )

            q_results[q_series_name] = series_to_agg.copy_override(
                expression=agg_expr,
                name=q_series_name,
            )

    quantiles_df = quantiles_df.copy_override(series=q_results)
    quantiles_df = quantiles_df.materialize(distinct=True)
    quantiles_df = _add_quantiles_as_series(quantiles_df, q)

    final_series = []
    for series_name in series_to_calculate:
        q_series_name = f'{series_name}_quantile'
        quantiles_df[q_series_name] = None
        quantiles_df[q_series_name] = quantiles_df[q_series_name].astype(df[series_name].dtype)
        for qt in quantiles:
            curr_q_series_name = f'__{series_name}_{qt}_quantile'
            mask = quantiles_df[_QUANTILES_SERIES_NAME] == qt

            current_q_series = quantiles_df[curr_q_series_name]

            if is_bigquery(df.engine) and isinstance(df[series_name], SeriesTimedelta):
                # cast back to SeriesTimeDelta
                current_q_series = SeriesTimedelta.from_total_seconds(
                    total_seconds=current_q_series.copy_override_type(SeriesFloat64)
                )

            quantiles_df.loc[mask, q_series_name] = current_q_series

        final_series.append(q_series_name)

    quantiles_df = quantiles_df.set_index(_QUANTILES_SERIES_NAME, append=True)
    return quantiles_df[final_series].materialize(node_name='quantile_calculation')


def _calculate_quantiles_with_linear_interpolation(
    df: DataFrame, q: Union[float, List[float]], group_by: List[str],
) -> DataFrame:
    """
    Helper function that simulates quantile calculation using linear interpolation based
    on method 7 of Hyndman & Fan.
    https://www.amherst.edu/media/view/129116/original/Sample+Quantiles.pdf

    Main reason to use this method is because Numpy's `quantile` and SQL's `percentile_cont` functions are
    based on it. Therefore, the expression generated by this function MUST yield exact results.

    Formula:
    𝑄(𝑝)=(1−𝜸)∗𝑋𝑗+𝜸∗𝑋𝑗+1

    where:
        𝒋  denotes the index of the element in the lowest boundary of the percentile.
        It is the intergal part of the "virtual index" (estimation of the linear interpolation).
        Which is defined as:
            𝒗𝒊𝒓𝒕𝒖𝒂𝒍_𝒊𝒏𝒅𝒆𝒙=𝑝∗𝑁+𝛼+𝑝∗(1−𝛼−𝛽)
            where:
                * p: percentile
                * N: Size of the population
                * 𝛼 = 1 and 𝛽 = 1 (by default, for linear interpolation)

            So, we can simplify it and express it as:
                𝒗𝒊𝒓𝒕𝒖𝒂𝒍_𝒊𝒏𝒅𝒆𝒙 = 𝑝∗𝑁+1+𝑝∗(1−1−1)
                           = 𝑝∗𝑁+1−𝑝
                           = 𝑝∗(𝑁−1)+1
            therefore,
                𝑗 = 𝑓𝑙𝑜𝑜𝑟(𝒗𝒊𝒓𝒕𝒖𝒂𝒍_𝒊𝒏𝒅𝒆𝒙)

        𝜸  is the interpolation parameter. Expressed as the fractional part of the virtual_index:
            𝛾 = 𝒗𝒊𝒓𝒕𝒖𝒂𝒍_𝒊𝒏𝒅𝒆𝒙%1

        𝑋𝑗 : element at position j in the sorted population:
            𝑋={𝑋1,...𝑋𝑗,𝑋𝑗+1,...,𝑋𝑛}

        𝑋𝑗+1 : adjacent element to  𝑋𝑗


    Numpy's Source
    calculation of virtual_index
        https://github.com/numpy/numpy/blob/54c52f13713f3d21795926ca4dbb27e16fada171/numpy/lib/function_base.py#L110

    calculation of gamma,  𝑋𝑗 ,  𝑋𝑗+1
        https://github.com/numpy/numpy/blob/54c52f13713f3d21795926ca4dbb27e16fada171/numpy/lib/function_base.py#L4704-L4713

    calculation of linear interpolation
        https://github.com/numpy/numpy/blob/54c52f13713f3d21795926ca4dbb27e16fada171/numpy/lib/function_base.py#L4513

    Supported only for Athena
    """
    if not is_athena(df.engine):
        raise Exception('Can only calculate quantiles using linear interpolation for Athena only.')

    quantiles_df = df.set_index(group_by, drop=True)

    series_to_calculate = _get_valid_series_for_calculation(df)

    # Step 1. aggregate all values of the numeric series into a sorted ARRAY

    transformed_series = []
    for series_name in series_to_calculate:
        series_to_agg = quantiles_df[series_name]
        current_window = (
            quantiles_df.sort_values(by=series_name).groupby(group_by)
            .window(end_boundary=WindowFrameBoundary.FOLLOWING).group_by
        )

        # mypy: help
        if not isinstance(current_window, Window):
            raise Exception('Invalid window object.')

        agg_expr = current_window.get_window_expression(
            AggregateFunctionExpression.construct('array_agg({})', series_to_agg)
        )

        grouped_series_name = f'__grouped_array_{series_name}'
        size_series_name = f'__size_{series_name}'
        transformed_series += [grouped_series_name, size_series_name]

        quantiles_df[grouped_series_name] = series_to_agg.copy_override(expression=agg_expr)
        quantiles_df[size_series_name] = series_to_agg.count(partition=current_window)

    quantiles_df = quantiles_df[transformed_series]
    quantiles_df = quantiles_df.materialize(distinct=True)

    # Step 2. add quantiles to calculate as const values in df
    quantiles_df = _add_quantiles_as_series(quantiles_df, q)
    quantiles_df = quantiles_df.set_index(_QUANTILES_SERIES_NAME, append=True)

    # Step 3. Calculate quantile based in linear interpolation
    # https://github.com/numpy/numpy/blob/54c52f13713f3d21795926ca4dbb27e16fada171/numpy/lib/function_base.py#L105-L112
    calculated_quantiles = {}
    for series_name in series_to_calculate:
        array_series = quantiles_df[f'__grouped_array_{series_name}']

        n_population = quantiles_df[f'__size_{series_name}']
        # (size of population - 1) * quantile + 1
        virtual_index = (
            (n_population - 1) * quantiles_df.all_series[_QUANTILES_SERIES_NAME] + 1
        )

        # gamma is the fractional part of the virtual_index (virtual_index % 1)
        gamma_series = virtual_index % 1

        # X_j: element located at floor(virtual_index) in the agg array
        X_j_series = gamma_series.copy_override(
            expression=Expression.construct(
                "try({}[cast({} as bigint)])", array_series, virtual_index // 1,
            )
        ).fillna(0.)

        # X_j_next: element after X_j
        X_j_next_series = gamma_series.copy_override(
            expression=Expression.construct(
                "try({}[cast({} as bigint) + 1])", array_series, virtual_index // 1,
            )
        ).fillna(0.)

        # final result
        # X_j + gamma * (X_j_next - X_j)
        result_series = X_j_series + gamma_series * (X_j_next_series - X_j_series)

        q_series_name = f'{series_name}_quantile'
        calculated_quantiles[q_series_name] = result_series.copy_override(name=f'{series_name}_quantile')
        if isinstance(df[series_name], SeriesTimedelta):
            calculated_quantiles[q_series_name] = SeriesTimedelta.from_total_seconds(
                calculated_quantiles[q_series_name].copy_override_type(SeriesFloat64)
            )

    quantiles_df = quantiles_df.copy_override(series=calculated_quantiles)
    return quantiles_df.materialize(node_name='quantile_calculation')


def calculate_quantiles_df(
    df: DataFrame,
    partition: WrappedPartition = None,
    q: Union[float, List[float]] = 0.5,
) -> DataFrame:
    """
    Calculates each requested quantile per each numeric/timedelta series contained in the DataFrame.

    Supports only quantile calculation based on linear interpolation.
    """
    if not _get_valid_series_for_calculation(df):
        raise ValueError('Cannot calculate quantiles, DataFrame has no numeric or timedelta series.')

    partition = partition or df.group_by
    if partition and not isinstance(partition, GroupBy):
        raise ValueError("DataFrame or provided partition is not valid. Expected GroupBy instance.")

    gb = list(partition.index.keys()) if partition else []

    df_cp = df.copy()
    if not gb:
        df_cp = df_cp.reset_index(drop=True)

    # removing current DataFrame's group_by, this way we avoid conflicts with partitioning
    df_cp = df_cp.copy_override(
        series={
            series.name: series.copy_override(group_by=None)
            for series in df_cp.data.values()
        },
        group_by=None
    )
    if not is_athena(df.engine):
        return _calculate_quantiles_with_percentile_cont(df_cp, q, group_by=gb)

    return _calculate_quantiles_with_linear_interpolation(df_cp, q, group_by=gb)



"""
Copyright 2021 Objectiv B.V.
"""
from typing import Union, Optional
from uuid import UUID

from sqlalchemy.engine import Dialect

from bach import DataFrameOrSeries
from bach.series import Series, value_to_series
from bach.expression import Expression
from bach.series.series import WrappedPartition, ToPandasInfo
from bach.types import StructuredDtype
from sql_models.constants import DBDialect
from sql_models.util import is_postgres, DatabaseNotSupportedException, is_bigquery, is_athena


class SeriesUuid(Series):
    """
    A Series that represents the UUID type and has UUID specific operations.

    **Database support and types**

    * Postgres: utilizes the native 'uuid' database type.
    * Athena: utilizes the generic 'varchar' database type.
    * BigQuery: utilizes the generic 'STRING' database type.
    """
    dtype = 'uuid'
    dtype_aliases = ()
    supported_db_dtype = {
        DBDialect.POSTGRES: 'uuid',
        # None here for BQ and Athena, because they don't have a uuid type.
        # We do support the UUID Series, but we store the data as a string data-type, which by default is
        # handled by SeriesString,
        DBDialect.ATHENA: None,
        DBDialect.BIGQUERY: None
    }

    supported_value_types = (UUID, str)
    supported_source_dtypes = ('string',)

    @classmethod
    def get_db_dtype(cls, dialect: Dialect) -> Optional[str]:
        if is_athena(dialect) or is_bigquery(dialect):
            from bach.series import SeriesString
            return SeriesString.get_db_dtype(dialect)
        return super().get_db_dtype(dialect)

    @classmethod
    def supported_literal_to_expression(cls, dialect: Dialect, literal: Expression) -> Expression:
        if is_postgres(dialect):
            return super().supported_literal_to_expression(dialect=dialect, literal=literal)
        if is_athena(dialect) or is_bigquery(dialect):
            from bach import SeriesString
            return SeriesString.supported_literal_to_expression(dialect=dialect, literal=literal)
        raise DatabaseNotSupportedException(dialect)

    @classmethod
    def supported_value_to_literal(
        cls,
        dialect: Dialect,
        value: Union[UUID, str],
        dtype: StructuredDtype
    ) -> Expression:
        if isinstance(value, str):
            # Check that the string value is a valid UUID by converting it to a UUID
            value = UUID(value)
        uuid_as_str = str(value)
        return Expression.string_value(uuid_as_str)

    @classmethod
    def dtype_to_expression(cls, dialect: Dialect, source_dtype: str, expression: Expression) -> Expression:
        if source_dtype == 'uuid':
            return expression
        if source_dtype in cls.supported_source_dtypes:
            if is_postgres(dialect):
                # If the format is wrong, then this will give an error later on, but there is not much we can
                # do about that here.
                return Expression.construct(f'cast({{}} as {cls.get_db_dtype(dialect)})', expression)
            if is_athena(dialect) or is_bigquery(dialect):
                return expression
            raise DatabaseNotSupportedException(dialect)
        # As far as we know the other types we support cannot be directly cast to uuid.
        raise ValueError(f'cannot convert {source_dtype} to uuid.')

    @classmethod
    def random(cls, base: DataFrameOrSeries) -> 'SeriesUuid':
        """
        Create a new Series object with an expression, that will evaluate to a random uuid for each row.

        :param base: DataFrame or Series from which the newly created Series' engine, base_node and index
            parameters are copied.

        .. warning::
            The returned Series has a non-deterministic expression, it will give a different result each
            time it is evaluated by the database.

        The non-deterministic expression can have some unexpected consequences. Consider the following code:

        .. code-block:: python

            df['x'] = SeriesUuid.random(df)
            df['y'] = df['x']
            df['different'] = df['y'] != df['x']

        The df['different'] series will be True for all rows, because the second statement copies the
        unevaluated expression, not the result of the expression. So at evaluation time the expression will
        be evaluated twice for each row, for the 'x' column and the 'y' column, giving different results both
        times. One way to work around this is to materialize the dataframe in its current state (using
        materialize()), before adding any columns that reference a column that's created with
        this function.
        """
        if is_postgres(base.engine):
            expr_str = 'gen_random_uuid()'
        elif is_athena(base.engine):
            # This works, which is interesting. Athena uses version 0.217 of prestodb [1], whose
            # documentation doesn't list this function [2]. The current version's docs do list it [3].
            # [1] https://docs.aws.amazon.com/athena/latest/ug/ddl-sql-reference.html
            # [2] https://prestodb.io/docs/0.217/index.html
            # [3] https://prestodb.io/docs/current/functions/uuid.html
            expr_str = 'uuid()'
        elif is_bigquery(base.engine):
            expr_str = 'GENERATE_UUID()'
        else:
            raise DatabaseNotSupportedException(base.engine)
        return cls.get_class_instance(
            engine=base.engine,
            base_node=base.base_node,
            index=base.index,
            name='random',
            expression=Expression.construct(expr_str),
            group_by=None,
            order_by=[],
            instance_dtype=cls.dtype
        )

    def to_pandas_info(self) -> Optional[ToPandasInfo]:
        if is_athena(self.engine) or is_bigquery(self.engine):
            return ToPandasInfo('object', lambda val: val if val is None else UUID(val))
        return None

    def _comparator_operation(self, other, comparator, other_dtypes=('uuid', 'string')):
        from bach import SeriesBoolean
        other = value_to_series(base=self, value=other)
        self_modified, other = self._get_supported(f"comparator '{comparator}'", other_dtypes, other)
        if other.dtype == 'string' and is_postgres(self.engine):
            expression = Expression.construct(
                f'({{}}) {comparator} (cast({{}} as uuid))', self_modified, other
            )
        else:  # other.dtype == 'uuid' or not is_postgres(self.engine)
            expression = Expression.construct(f'({{}}) {comparator} ({{}})', self_modified, other)
        return self_modified.copy_override_type(SeriesBoolean).copy_override(expression=expression)

    def min(self, partition: WrappedPartition = None, skipna: bool = True):
        """ INTERNAL: Only here to not trigger errors from describe """
        raise NotImplementedError()

    def max(self, partition: WrappedPartition = None, skipna: bool = True):
        """ INTERNAL: Only here to not trigger errors from describe """
        raise NotImplementedError()



"""
Copyright 2022 Objectiv B.V.
"""
from abc import abstractmethod

from sqlalchemy.engine import Dialect

from bach.dataframe import DataFrameOrSeries
import itertools
from collections import defaultdict
from typing import Dict, Hashable, List, Set, Sequence, TypeVar, Generic

from bach.dataframe import DtypeNamePair
from bach import DataFrame, get_series_type_from_dtype, Series
from bach.expression import Expression, join_expressions
from bach.sql_model import BachSqlModel, construct_references
from bach.utils import ResultSeries, get_result_series_dtype_mapping, get_merged_series_dtype, \
    get_sql_column_name, get_name_to_column_mapping
from sql_models.model import CustomSqlModelBuilder, Materialization

TDataFrameOrSeries = TypeVar('TDataFrameOrSeries', bound='DataFrameOrSeries')

DEFAULT_CONCAT_SERIES_DTYPE = 'string'


class ConcatOperation(Generic[TDataFrameOrSeries]):
    """
    Abstract class that specifies the list of objects to be concatenated.

    Child classes are in charge of specifying the correct type (DataFrame/Series) of all objects.
    All classes should implement:
     - _get_concatenated_object method: returns a single object with the correct instantiated type.
     - _join_series_expressions method: returns expression per object to be concatenated
     - _get_series method: returns the data series of the final concatenated object
    """
    objects: Sequence[TDataFrameOrSeries]
    ignore_index: bool

    def __init__(self, objects: Sequence[TDataFrameOrSeries], ignore_index: bool = False):
        self.objects = objects
        self.ignore_index = ignore_index

    def __call__(self, *args, **kwargs) -> TDataFrameOrSeries:
        """
        If only one object is provided for concatenation, it should just return a copy.
        Otherwise, child class _get_concatenated_object method is called.
        """
        if not len(self.objects):
            raise ValueError('no objects to concatenate.')

        if len(self.objects) == 1:
            index = {} if self.ignore_index else self.objects[0].index
            return self.objects[0].copy_override(index=index)  # type: ignore

        return self._get_concatenated_object()

    @property
    def dialect(self) -> Dialect:
        return self.objects[0].engine.dialect

    def _get_indexes(self) -> Dict[str, ResultSeries]:
        """
        gets the indexes of the final concatenated dataframe or series.
        All objects must have the same indexes, otherwise concatenation is not performed.
        """
        from bach.series import SeriesAbstractMultiLevel
        if self.ignore_index:
            return {}

        all_indexes = list(itertools.chain.from_iterable(obj.index.values() for obj in self.objects))
        merged_indexes = self._get_result_series(all_indexes)

        # all objects should have the same indexes
        for obj in self.objects:
            curr_obj_index = list(itertools.chain.from_iterable(
                [idx] if not isinstance(idx, SeriesAbstractMultiLevel) else list(idx.levels.values())
                for idx in obj.index.values()
            ))
            if set(merged_indexes.keys()) != {idx.name for idx in curr_obj_index}:
                raise ValueError('concatenation with different index levels is not supported yet.')

            if any(mi.dtype != oi.dtype for mi, oi in zip(merged_indexes.values(), curr_obj_index)):
                raise ValueError('concatenation with different index dtypes is not supported yet.')

        return merged_indexes

    def _get_result_series(self, series: List[Series]) -> Dict[str, ResultSeries]:
        """
        merges all shared series and defines final dtype
        """
        from bach.series import SeriesAbstractMultiLevel
        series_dtypes = defaultdict(set)
        for s in series:
            if not isinstance(s, SeriesAbstractMultiLevel):
                series_dtypes[s.name].add(s.dtype)
                continue

            for lvl in s.levels.values():
                series_dtypes[lvl.name].add(lvl.dtype)

        series_names = series_dtypes.keys()

        return {
            series_name: ResultSeries(
                name=series_name,
                expression=Expression.identifier(series_name),
                dtype=get_merged_series_dtype(series_dtypes[series_name]),
            )
            for series_name in series_names
        }

    def _get_model(
        self,
        objects: Sequence[TDataFrameOrSeries],
        variables: Dict['DtypeNamePair', Hashable],
    ) -> 'ConcatSqlModel':
        """
        returns a ConcatSqlModel which unifies all queries from all dataframes.
        """
        new_indexes = self._get_indexes()
        new_data_series = self._get_series()

        series_expressions = [self._join_series_expressions(df) for df in objects]

        new_columns_rs = list(new_indexes.values()) + list(new_data_series.values())
        column_expressions = {rs.name: rs.expression for rs in new_columns_rs}

        return ConcatSqlModel.get_instance(
            dialect=self.dialect,
            column_expressions=column_expressions,
            all_series_expressions=series_expressions,
            all_nodes=[df.base_node for df in objects],
            variables=variables,
        )

    @abstractmethod
    def _get_concatenated_object(self) -> TDataFrameOrSeries:
        """
        returns a single object based on the instantiated type from the child class
        """
        raise NotImplementedError()

    @abstractmethod
    def _join_series_expressions(self, obj: TDataFrameOrSeries) -> Expression:
        """
        return the column expression for the object subquery. For more information see implementation.
        """
        raise NotImplementedError()

    @abstractmethod
    def _get_series(self) -> Dict[str, ResultSeries]:
        """
        gets the data series of the final concatenated object
        """
        raise NotImplementedError()


class DataFrameConcatOperation(ConcatOperation[DataFrame]):
    """
    In order to instantiate this class you should provide the following params:
    objects: a list of DataFrames to be concatenated (all DataFrames should have the same indexes)
    ignore_index: a boolean specifying if the resultant DataFrame must preserve the original indexes or not.
    sort: a boolean specifying the order of the data_columns in the result. If False, the order of series
    will be based on each object's position and data_columns value.

    returns a new DataFrame

    Example:
        DataFrameConcatOperation(objects=[df1, df2], ignore_index=True, sort=True)()
    """
    sort: bool

    def __init__(self, objects: Sequence[DataFrame], ignore_index: bool = False, sort: bool = False):
        self.sort = sort
        super().__init__(objects=objects, ignore_index=ignore_index)

    def _get_overridden_objects(self) -> List[DataFrame]:
        """
        generates a copy for each dataframe and prepares them for concatenation
        """
        dfs: List[DataFrame] = []
        for obj in self.objects:
            if isinstance(obj, Series):
                raise Exception('Cannot concat Series to DataFrame')

            df = obj.copy() if not self.ignore_index else obj.reset_index(drop=True)
            # need to materialize in order to avoid further problems
            if df.group_by:
                df = df.materialize()

            dfs.append(df)

        return dfs

    def _join_series_expressions(self, obj: DataFrame) -> Expression:
        """
        generates the column expression for the object subquery
        - if a column doesn't exist in the object's base_node, all values will be null
        - if the column exists but has a different dtype from the result, it will be casted
        """
        new_indexes = self._get_indexes()
        new_data_series = self._get_series()

        expressions = []

        all_result_series: Dict[str, ResultSeries] = {**new_indexes, **new_data_series}
        for idx, rc in all_result_series.items():
            if idx not in obj.all_series:
                null_expr = Expression.raw('NULL')
                expressions.append(
                    Expression.construct_expr_as_sql_name(dialect=self.dialect, expr=null_expr, name=rc.name)
                )
                continue

            has_diff_dtype = rc.dtype != obj.all_series[idx].dtype
            curr_series = obj.all_series[idx] if not has_diff_dtype else obj.all_series[idx].astype(rc.dtype)
            expressions.append(
                Expression.construct_expr_as_sql_name(
                    dialect=self.dialect,
                    expr=curr_series.expression,
                    name=rc.name
                )
            )

        return join_expressions(expressions)

    def _get_series(self) -> Dict[str, ResultSeries]:
        """
        gets the data series of the final concatenated dataframe
        """
        all_series = list(
            itertools.chain.from_iterable(df.data.values() for df in self.objects)
        )
        result_series = self._get_result_series(all_series)
        if not self.sort:
            return result_series

        return {s: result_series[s] for s in sorted(result_series)}

    def _get_concatenated_object(self) -> DataFrame:
        """
        prepares all dataframes for concatenation (all of them have the same columns)
        gets the resultant indexes and series for the final dataframe.

        returns a new dataframe with all the rows from the provided objects
        """
        objects = self._get_overridden_objects()
        new_indexes = self._get_indexes()
        new_data_series = self._get_series()

        main_df: DataFrame = objects[0]
        variables = {}
        savepoints = main_df.savepoints

        # variables are overridden based on the order of concatenation. This means that the initial dataframe
        # will have higher priority over the following dataframes
        for df in reversed(objects):
            variables.update(df.variables)
            savepoints.merge(df.savepoints)

        new_columns_rs = list(new_indexes.values()) + list(new_data_series.values())
        name_to_column_mapping = get_name_to_column_mapping(self.dialect, [rs.name for rs in new_columns_rs])
        return main_df.copy_override(
            base_node=self._get_model(objects, variables),
            index_dtypes=get_result_series_dtype_mapping(list(new_indexes.values())),
            series_dtypes=get_result_series_dtype_mapping(list(new_data_series.values())),
            savepoints=savepoints,
            variables=variables,
            name_to_column_mapping=name_to_column_mapping
        )


class SeriesConcatOperation(ConcatOperation[Series]):
    """
    In order to instantiate this class you should provide the following params:
    objects: a list of Series to be concatenated (all Series should have the same indexes)
    ignore_index: a boolean specifying if the resultant Series must preserve the original indexes or not.

    returns a new Series

    Example:
        SeriesConcatOperation(objects=[s1, s2], ignore_index=True)()
    """
    def _get_overridden_objects(self) -> List[Series]:
        """
        creates new copies for each series to be concatenated
        """
        series: List[Series] = []
        for obj in self.objects:
            if isinstance(obj, DataFrame):
                raise Exception('Cannot concat DataFrame to Series')

            df = obj.to_frame()
            if df.group_by:
                df = df.materialize()
            series.append(df.all_series[obj.name])

        return series

    def _get_series(self) -> Dict[str, ResultSeries]:
        """
        gets the final data series result
        """
        dtypes: Set[str] = set(series.dtype for series in self.objects)

        main_series = self.objects[0].copy_override_dtype(dtype=get_merged_series_dtype(dtypes))
        return self._get_result_series([main_series])

    def _join_series_expressions(self, obj: Series) -> Expression:
        """
        generates the column expression for the object subquery
        """
        result_series = list(self._get_series().values())[0]

        series_expression = Expression.construct_expr_as_sql_name(
            dialect=self.dialect,
            expr=obj.astype(result_series.dtype).expression,
            name=result_series.name,
        )

        if self.ignore_index:
            return series_expression

        index_expressions = [idx.expression for idx in self._get_indexes().values()]
        return join_expressions(index_expressions + [series_expression])

    def _get_concatenated_object(self) -> Series:
        """
        returns a series that contains all rows from all provided objects.
        """
        objects = self._get_overridden_objects()
        main_series: Series = objects[0]
        final_result_series: ResultSeries = list(self._get_series().values())[0]

        series_cls = get_series_type_from_dtype(final_result_series.dtype)
        node = self._get_model(objects, variables={})
        index = {
            idx_name: idx.copy_override(base_node=node)
            for idx_name, idx in main_series.index.items()
        }
        sql_column_name = get_sql_column_name(dialect=self.dialect, name=final_result_series.name)
        return series_cls(
            engine=main_series.engine,
            base_node=node,
            name=final_result_series.name,
            expression=Expression.column_reference(sql_column_name),
            index={} if self.ignore_index else index,
            group_by=None,
            order_by=[],
            instance_dtype=series_cls.dtype  # TODO: make this work for structural types too
        )


class ConcatSqlModel(BachSqlModel):
    @classmethod
    def get_instance(
        cls,
        *,
        dialect: Dialect,
        column_expressions: Dict[str, Expression],
        all_series_expressions: List[Expression],
        all_nodes: List[BachSqlModel],
        variables: Dict['DtypeNamePair', Hashable],
    ) -> 'ConcatSqlModel':
        name = 'concat_sql'
        base_sql = 'select {serie_expr} from {node}'
        sql = ' union all '.join(
            base_sql.format(serie_expr=col_expr.to_sql(dialect), node=f"{{{{node_{idx}}}}}")
            for idx, col_expr in enumerate(all_series_expressions)
        )

        references = construct_references(
            base_references={f'node_{idx}': node for idx, node in enumerate(all_nodes)},
            expressions=all_series_expressions
        )

        return ConcatSqlModel(
            model_spec=CustomSqlModelBuilder(sql=sql, name=name),
            placeholders=cls._get_placeholders(dialect, variables, all_series_expressions),
            references=references,
            materialization=Materialization.CTE,
            materialization_name=None,
            column_expressions=column_expressions,
        )



"""
Copyright 2021 Objectiv B.V.
"""
import re
from typing import NamedTuple, Dict, List, Union, cast

from sqlalchemy.engine import Engine, Dialect

from bach import DataFrame
from bach.sql_model import BachSqlModel
from sql_models.model import Materialization, SqlModel, CustomSqlModelBuilder
from sql_models.sql_generator import to_sql_materialized_nodes, GeneratedSqlStatement
from sql_models.util import quote_identifier


class SavepointEntry(NamedTuple):
    """
    Class to represent a savepoint
    """
    name: str
    df_original: 'DataFrame'
    materialization: Materialization


class CreatedObject(NamedTuple):
    """
    Class representing a created database object (view/table)
    """
    name: str
    materialization: Materialization


class Savepoints:
    """
    Class to store a collection of savepoints. A savepoint represents the state of a DataFrame instance at
    the moment it is being added to a Savepoints object.

    Functionality:
    - Convert the savepoints to SQL using :meth:`Savepoints.to_sql()`
    - Materialize as tables or views in a database using :meth:`Savepoints.write_to_databse()`.
    - TODO: serialize to a file, and later restored by deserializing
    - TODO: export as a DBT project
    - TODO: export to BI tools
    """

    def __init__(self):
        self._entries: Dict[str, SavepointEntry] = {}

    @property
    def all(self) -> List[SavepointEntry]:
        return list(self._entries.values())

    @property
    def names(self) -> List[str]:
        return [entry.name for entry in self._entries.values()]

    def merge(self, other: 'Savepoints'):
        """
        INTERNAL
        Update this Savepoints object by adding all savepoints from other.
        """
        for name, entry in other._entries.items():
            if name in self._entries:
                existing = self._entries[name]
                if existing.df_original != entry.df_original \
                        or existing.materialization != entry.materialization:
                    raise ValueError(f'Conflicting savepoints. The savepoint "{name}" exists in both '
                                     f'Savepoints objects, but is different.')
            else:
                self._entries[name] = SavepointEntry(
                    name=name,
                    df_original=entry.df_original.copy(),
                    materialization=entry.materialization,
                )

    def add_savepoint(self, name: str, df: DataFrame, materialization: Materialization):
        """
        Add the DataFrame as a savepoint.

        Generally one would use :py:meth:`bach.DataFrame.set_savepoint()`
        """
        if name is None or not re.match('^[a-zA-Z0-9_]+$', name):
            raise ValueError(f'Name must match ^[a-zA-Z0-9_]+$, name: "{name}"')
        if name in self._entries:
            existing = self._entries[name]
            if existing.df_original != df or existing.materialization != materialization:
                raise ValueError(f'A different savepoint with the name "{name}" already exists.')
            # Nothing to do, we already have this entry
            return
        if not df.is_materialized:
            raise ValueError('Can only create a savepoint from materialized DataFrames.')
        existing_basenodes = {entry.df_original.base_node: entry.name for entry in self._entries.values()}
        if df.base_node in existing_basenodes:
            # We cannot support having the same base_node as multiple savepoints. This will give problems
            # if the different savepoints have different materialization
            raise ValueError(f'Another savepoint has the same base_node, not allowed. Other savepont: '
                             f'{existing_basenodes[df.base_node]}')

        self._entries[name] = SavepointEntry(
            name=name,
            df_original=df.copy(),
            materialization=materialization
        )

    def set_materialization(self, name: str, materialization: Union[Materialization, str]):
        """
        Set the materialization of a savepoint, prior to calling :meth:`write_to_db()`. Calling this function
        does not materialize anything to the database by itself.

        NOTE: This alos does not undo any side-effects from earlier called functions, such as
            :meth:`write_to_db()` i.e. if materialization was 'table' before and is changed to 'view', then
            an existing table in the database is not updated into a view.

        :param name: name of savepoint
        :param materialization: materialization value
        """
        materialization = Materialization.normalize(materialization)
        current = self._entries[name]
        if current.materialization == materialization:
            return
        self._entries[name] = SavepointEntry(
            name=current.name,
            df_original=current.df_original,
            materialization=materialization
        )

    def remove_savepoint(self, name: str):
        """
        Discard the savepoint.
        NOTE: This does not undo any side-effects from earlier called function, such as :meth:`write_to_db()`
        """
        del self._entries[name]

    def get_df(self, savepoint_name: str) -> 'DataFrame':
        """
        Return a copy of the original DataFrame that was saved with the given name.
        """
        return self._entries[savepoint_name].df_original.copy()

    def get_materialized_df(self, savepoint_name: str, engine_override: Engine = None) -> 'DataFrame':
        """
        Return the DataFrame that was saved with the given name, but with an updated base_node.
        The updated base_node assumes that :meth:`write_to_db()` has been executed. Where possible it
        will query one of the created tables or views, instead of the original source tables.
        :param savepoint_name: name of the savepoint
        :param engine_override: optional, allows overriding the 'engine' of the returned dataframe. This is
            useful if the savepoints were persisted to a different database than from which they originally
            came (typically if write_to_db() was invoked with a different `engine` argument than the
            original_df's engine).
        """
        info = self._entries[savepoint_name]
        full_graph = self._get_combined_graph()
        graph = full_graph.references[f'ref_{info.name}']
        bach_graph = cast(BachSqlModel, graph)  # help mypy, see _get_virtual_node() for why this is correct
        new_df = info.df_original.copy_override_base_node(base_node=bach_graph)
        if engine_override is not None:
            new_df = new_df.copy_override(engine=engine_override)
        return new_df

    def write_to_db(self, engine_override: Engine = None, overwrite: bool = False) -> List[CreatedObject]:
        """
        Create the tables and views for all of the savepoints that have a table or view materialization.

        :param engine_override: optional. If not set this will use the engine of the original dataframes in
            the savepoints. If the savepoints do not all share the same engine, then this parameter is
            mandatory.
        :param overwrite: If true, drop table/view statements will be run first
        :return: List of all created objects
        """
        if not self._entries:
            return []  # nothing to do
        if engine_override:
            engine = engine_override
        else:
            engines = {entry.df_original.engine for entry in self._entries.values()}
            if len(engines) != 1:
                raise ValueError("engine_override cannot be None if the savepoints's entries don't all "
                                 "share the same engine.")
            engine = list(engines)[0]
        result_created = []
        drop_statements = self.get_drop_statements(dialect=engine.dialect)
        create_statements = self.get_create_statements(dialect=engine.dialect)

        with engine.connect() as conn:
            with conn.begin() as transaction:
                if overwrite:
                    # This is a bit fragile. Drop statements might fail if other objects (which we might not
                    # consider) depend on a view/table, or if the object type (view/table) is different than
                    # we assume. For now that's just the way it is, the user will get an error.
                    if drop_statements:
                        drop_sql = '; '.join(drop_statements.values())
                        conn.execute(drop_sql)

                for name, statement in create_statements.items():
                    info = self._entries[name]
                    conn.execute(statement)
                    result_created.append(CreatedObject(name=name, materialization=info.materialization))
                transaction.commit()
        return result_created

    def get_drop_statements(self, dialect: Dialect) -> Dict[str, str]:
        """
        Get the drop statements to remove all savepoints that are marked as table or view.
        The returned dictionary is sorted, such that depending tables/views are deleted before dependencies.
        :param dialect: SQL Dialect
        :return: dict with as key the savepoint name, and as value the drop statement for that table/view
        """
        sql_statements = self.to_sql(dialect)
        drop_statements = {}
        for sql_stat in reversed(sql_statements):
            name = sql_stat.name
            if sql_stat.materialization == Materialization.TABLE:
                drop_statements[name] = f'drop table if exists {quote_identifier(dialect, name)}'
            elif sql_stat.materialization == Materialization.VIEW:
                drop_statements[name] = f'drop view if exists {quote_identifier(dialect, name)}'
        return drop_statements

    def get_create_statements(self, dialect: Dialect) -> Dict[str, str]:
        """
        Get the create statements to create savepoints that are marked as table or view.
        The returned dictionary is sorted, such that dependencies are created before tables/views that
        depend on them.
        :param dialect: SQL Dialect
        :return: dict with as key the savepoint name, and as value the create statement for that table/view
        """
        sql_statements = self.to_sql(dialect)
        return {
            sql_stat.name: sql_stat.sql for sql_stat in sql_statements
            if sql_stat.materialization in (Materialization.TABLE, Materialization.VIEW)
        }

    def to_sql(self, dialect: Dialect) -> List[GeneratedSqlStatement]:
        """
        Generate the sql for all save-points
        :param dialect: SQL Dialect
        :return: List of GeneratedSqlStatement, each representing one savepoint
        """
        graph = self._get_combined_graph()
        sqls = to_sql_materialized_nodes(dialect=dialect, start_node=graph, include_start_node=False)
        return sqls

    def _get_combined_graph(self) -> SqlModel:
        """
        Get a single graph that contains all savepoints.

        The savepoints are referred by the returned sql-model as 'ref_{name}'.
        """
        entries = list(self._entries.values())
        references: Dict[str, BachSqlModel] = {
            f'ref_{entry.name}': entry.df_original.base_node for entry in entries
        }
        # Create one graph with all entries
        graph = _get_virtual_node(references)

        # Now update all the nodes that represent an entry, to have the correct materialization
        for entry in entries:
            reference_path = (f'ref_{entry.name}', )
            graph = graph.set_materialization_name(reference_path, materialization_name=entry.name)
            graph = graph.set_materialization(reference_path, materialization=entry.materialization)
        return graph

    def __str__(self) -> str:
        """ Give string with overview of all savepoints per materialization. """
        tables = [entry for entry in self.all if entry.materialization == Materialization.TABLE]
        views = [entry for entry in self.all if entry.materialization == Materialization.VIEW]
        others = [entry for entry in self.all if entry.materialization
                  not in (Materialization.TABLE, Materialization.VIEW)]
        result = [f'Savepoint, entries: {len(self.all)}']

        result.append(f'\tTables, entries: {len(tables)}')
        for table in tables:
            result.append(f'\t\t{table.name}')

        result.append(f'\tViews, entries: {len(views)}')
        for view in views:
            result.append(f'\t\t{view.name}')

        result.append(f'\tOther, entries: {len(others)}')
        for other in others:
            result.append(f'\t\t{other.name}')

        string = '\n'.join(result)
        return string


def _get_virtual_node(references: Dict[str, BachSqlModel]) -> SqlModel:
    # TODO: move this to sqlmodel?
    # reference_sql is of form "{{ref_0}}, {{1}}, ..., {{n}}"
    reference_sql = ', '.join(f'{{{{{ref_name}}}}}' for ref_name in references.keys())
    sql = f'select * from {reference_sql}'
    return CustomSqlModelBuilder(name='virtual_node', sql=sql)\
        .set_materialization(Materialization.VIRTUAL_NODE)\
        .set_values(**references)\
        .instantiate()
